from typing import Any, List, Tuple

import pytest
from pydantic import BaseModel

from ..mathy import envs
from ..mathy.agents.policy_value_model import PolicyValueModel
from ..mathy.agents.zero.config import SelfPlayConfig
from ..mathy.agents.zero.practice_runner import PracticeRunner
from ..mathy.agents.zero.trainer import SelfPlayTrainer
from ..mathy.env import MathyEnv
from ..mathy.state import MathyObservation, observations_to_window


def test_mathy_zero_trainer_constructor():
    config = SelfPlayConfig()
    env: MathyEnv = envs.PolySimplify()
    model = PolicyValueModel(config, predictions=env.action_size)
    assert SelfPlayTrainer(config, model, env.action_size) is not None


def test_mathy_zero_practice_runner():
    with pytest.raises(ValueError):
        # Must be SelfPlayConfig or a subclass
        PracticeRunner(BaseModel())

    runner = PracticeRunner(SelfPlayConfig())
    with pytest.raises(NotImplementedError):
        runner.get_env()
    with pytest.raises(NotImplementedError):
        runner.get_model(None)


def test_mathy_zero_practice_runner_execute_episode():
    runner = PracticeRunner(SelfPlayConfig())

    class FakeEnv:
        def __init__(self):
            self.state = None

        def reset(self):
            pass

    with pytest.raises(ValueError):
        runner.execute_episode(1, None, None, "", False)
    with pytest.raises(ValueError):
        runner.execute_episode(1, FakeEnv(), None, "", False)
    with pytest.raises(ValueError):
        runner.execute_episode(1, FakeEnv(), FakeEnv(), "", False)



from typing import Any, Dict, List, Optional, Tuple, Union
import tensorflow as tf

from mathy.agents.base_config import BaseConfig
from mathy.core.expressions import MathTypeKeysMax
from mathy.state import (
    MathyInputsType,
    MathyWindowObservation,
    ObservationFeatureIndices,
)
from mathy.agents.densenet import DenseNetStack

from .attention import SeqSelfAttention


class MathyEmbedding(tf.keras.Model):
    def __init__(self, config: BaseConfig, **kwargs):
        super(MathyEmbedding, self).__init__(**kwargs)
        self.config = config
        self.token_embedding = tf.keras.layers.Embedding(
            input_dim=MathTypeKeysMax,
            output_dim=self.config.embedding_units,
            name="nodes_input",
            mask_zero=True,
        )
        # +1 for the value
        # +2 for the problem type hashes
        self.concat_size = 3
        self.values_dense = tf.keras.layers.Dense(
            self.config.units, name="values_input"
        )
        self.type_dense = tf.keras.layers.Dense(self.config.units, name="type_input")
        self.in_dense = tf.keras.layers.Dense(
            self.config.units,
            # In transform gets the embeddings concatenated with the
            # floating point value at each node.
            input_shape=(None, self.config.embedding_units + self.concat_size),
            name="in_dense",
            activation="relu",
            kernel_initializer="he_normal",
        )
        self.output_dense = tf.keras.layers.Dense(
            self.config.embedding_units,
            name="out_dense",
            activation="relu",
            kernel_initializer="he_normal",
        )

        NormalizeClass = tf.keras.layers.LayerNormalization
        if self.config.normalization_style == "batch":
            NormalizeClass = tf.keras.layers.BatchNormalization
        self.out_dense_norm = NormalizeClass(name="out_dense_norm")
        self.nodes_lstm_norm = NormalizeClass(name="lstm_nodes_norm")
        self.time_lstm_norm = NormalizeClass(name="time_lstm_norm")
        self.lstm_nodes = tf.keras.layers.Bidirectional(
            tf.keras.layers.LSTM(
                self.config.lstm_units,
                name="nodes_lstm",
                time_major=False,
                return_sequences=True,
                dropout=self.config.dropout,
            ),
            merge_mode="ave",
        )
        self.lstm_time = tf.keras.layers.LSTM(
            self.config.lstm_units,
            name="time_lstm",
            time_major=True,
            return_sequences=True,
            dropout=self.config.dropout,
        )
        self.lstm_attention = SeqSelfAttention(
            attention_activation="sigmoid",
            name="self_attention",
            return_attention=True,
        )

    def call(self, features: MathyInputsType, train: tf.Tensor = None) -> tf.Tensor:
        output = tf.concat(
            [
                self.token_embedding(features[ObservationFeatureIndices.nodes]),
                self.values_dense(
                    tf.expand_dims(features[ObservationFeatureIndices.values], axis=-1)
                ),
                self.type_dense(features[ObservationFeatureIndices.type]),
            ],
            axis=-1,
            name="input_vectors",
        )
        output = self.in_dense(output)
        output = self.lstm_time(output)
        output = self.time_lstm_norm(output)
        output = self.lstm_nodes(output)
        output = self.nodes_lstm_norm(output)
        output, attention = self.lstm_attention(output)
        return self.out_dense_norm(self.output_dense(output)), attention



import json
import os
from queue import Queue
from typing import List

import gym
import numpy as np

from ...teacher import Teacher
from ..policy_value_model import get_or_create_policy_model, PolicyValueModel
from .config import A3CConfig
from .worker import A3CWorker
from ...envs.gym import MathyGymEnv
from ...state import observations_to_window


class A3CAgent:

    args: A3CConfig
    global_model: PolicyValueModel

    def __init__(self, args: A3CConfig, env_extra: dict = None):
        import tensorflow as tf

        self.args = args
        self.env_extra = env_extra if env_extra is not None else {}
        if self.args.verbose:
            print(f"Agent: {os.path.join(args.model_dir, args.model_name)}")
            print(f"Config: {json.dumps(self.args.dict(), indent=2)}")
        self.teacher = Teacher(
            topic_names=self.args.topics,
            num_students=self.args.num_workers,
            difficulty=self.args.difficulty,
            eval_window=self.args.teacher_evaluation_steps,
            win_threshold=self.args.teacher_promote_wins,
            lose_threshold=self.args.teacher_demote_wins,
        )
        env: MathyGymEnv = gym.make(self.teacher.get_env(0, 0), **self.env_extra)
        self.action_size = env.action_space.n
        self.log_dir = os.path.join(self.args.model_dir, "tensorboard")
        self.writer = tf.summary.create_file_writer(self.log_dir)
        initial_window = observations_to_window([env.reset()])
        self.global_model = get_or_create_policy_model(
            args=args, predictions=self.action_size, is_main=True, env=env.mathy
        )
        with self.writer.as_default():
            tf.summary.trace_on(graph=True)
            inputs = initial_window.to_inputs()

            @tf.function
            def trace_fn():
                return self.global_model.call(inputs)

            trace_fn()
            tf.summary.trace_export(
                name="PolicyValueModel", step=0, profiler_outdir=self.log_dir
            )
            tf.summary.trace_off()
            if self.args.verbose:
                print(self.global_model.summary())

    def train(self):
        res_queue = Queue()
        exp_out_queue = Queue()
        A3CWorker.global_episode = 0
        worker_exploration_epsilons = np.geomspace(
            self.args.e_greedy_min, self.args.e_greedy_max, self.args.num_workers
        )
        workers = [
            A3CWorker(
                env_extra=self.env_extra,
                global_model=self.global_model,
                action_size=self.action_size,
                greedy_epsilon=worker_exploration_epsilons[i],
                args=self.args,
                teacher=self.teacher,
                worker_idx=i,
                optimizer=self.global_model.optimizer,
                result_queue=res_queue,
                writer=self.writer,
            )
            for i in range(self.args.num_workers)
        ]

        for i, worker in enumerate(workers):
            worker.start()

        try:
            while True:
                reward = res_queue.get()
                if reward is None:
                    break
        except KeyboardInterrupt:
            print("Received Keyboard Interrupt. Shutting down.")
            A3CWorker.request_quit = True

        # Do an optimistic save incase there's a problem joining the workers
        self.global_model.save()
        [w.join() for w in workers]
        # Do a final save after joining to get the very latest model
        self.global_model.save()
        print("Done. Bye!")




# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""TensorFlow ops for discrete-action Policy Gradient functions."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

# Dependency imports
from six.moves import zip
import tensorflow as tf
from . import base_ops
from . import value_ops
from tensorflow.python.util import nest

DiscretePolicyEntropyExtra = collections.namedtuple(
    "discrete_policy_entropy_extra", ["entropy"]
)
SequenceAdvantageActorCriticExtra = collections.namedtuple(
    "sequence_advantage_actor_critic_extra",
    [
        "entropy",
        "entropy_loss",
        "baseline_loss",
        "policy_gradient_loss",
        "advantages",
        "discounted_returns",
    ],
)


def discrete_policy_entropy_loss(
    policy_logits, normalise=False, name="discrete_policy_entropy_loss"
):
    """Computes the entropy 'loss' for a batch of policy logits.

  Given a batch of policy logits, calculates the entropy and corrects the sign
  so that minimizing the resulting loss op is equivalent to increasing entropy
  in the batch. This loss is optionally normalised to the range `[-1, 0]` by
  dividing by the log number of actions. This makes it more invariant to the
  size of the action space.

  This function accepts a nested array of `policy_logits` in order
  to allow for multiple discrete actions. In this case, the loss is given by
  `-sum_i(H(p_i))` where `p_i` are members of the `policy_logits` nest and
  H is the Shannon entropy.

  Args:
    policy_logits: A (possibly nested structure of) (N+1)-D Tensor(s) with
        shape `[..., A]`,  representing the log-probabilities of a set of
        Categorical distributions, where `...` represents at least one
        dimension (e.g., batch, sequence), and `A` is the number of discrete
        actions (which need not be identical across all tensors).
        Does not need to be centered.
    normalise: If True, divide the loss by the `sum_i(log(A_i))` where `A_i`
        is the number of actions for the i'th tensor in the `policy_logits`
        nest. Default is False.
    name: Optional, name of this op.

  Returns:
    A namedtuple with fields:

    * `loss`: Entropy 'loss', shape `[B]`.
    * `extra`: a namedtuple with fields:
        * `entropy`: Entropy of the policy, shape `[B]`.
  """
    policy_logits = nest.flatten(policy_logits)

    with tf.compat.v1.name_scope(name, values=policy_logits):
        entropy = tf.add_n(
            [
                tf.reduce_sum(
                    input_tensor=-tf.nn.softmax(scalar_policy_logits)
                    * tf.nn.log_softmax(scalar_policy_logits),
                    axis=-1,
                )
                for scalar_policy_logits in policy_logits
            ],
            name="entropy",
        )
        # We want a value that we can minimize along with other losses, and where
        # minimizing means driving the policy towards a uniform distribution over
        # the actions. We thus scale it by negative one so that it can be simply
        # added to other losses.
        scale = tf.constant(-1.0, dtype=tf.float32)
        if normalise:
            num_actions = [
                tf.cast(tf.shape(input=scalar_policy_logits)[-1], dtype=tf.float32)
                for scalar_policy_logits in policy_logits
            ]
            scale /= tf.reduce_sum(input_tensor=tf.math.log(tf.stack(num_actions)))
        loss = tf.multiply(scale, entropy, name="entropy_loss")

    return base_ops.LossOutput(loss, DiscretePolicyEntropyExtra(entropy))


def sequence_advantage_actor_critic_loss(
    policy_logits,
    baseline_values,
    actions,
    rewards,
    pcontinues,
    bootstrap_value,
    lambda_=1,
    entropy_cost=None,
    baseline_cost=1,
    normalise_entropy=False,
    name="SequenceAdvantageActorCriticLoss",
):
    """Calculates the loss for an A2C update along a batch of trajectories.

  Technically A2C is the special case where lambda=1; for general lambda
  this is the loss for Generalized Advantage Estimation (GAE), modulo chunking
  behaviour if passing chunks of episodes (see `generalized_lambda_returns` for
  more detail).

  Note: This function takes policy _logits_ as input, not the log-policy like
  Reinforce does.

  This loss jointly learns the policy and the baseline. Therefore, gradients
  for this loss flow through each tensor in `policy_logits` and
  `baseline_values`, but no other input tensors. The policy is learnt with the
  advantage actor-critic loss, plus an optional entropy term. The baseline is
  regressed towards the n-step bootstrapped returns given by the
  reward/pcontinue sequence.  The `baseline_cost` parameter scales the
  gradients w.r.t the baseline relative to the policy gradient. i.e:
  `d(loss) / d(baseline) = baseline_cost * (n_step_return - baseline)`.

  `rewards` and `pcontinues` are the sequences of data taken directly from the
  environment, possibly modulated by a discount. `baseline_values` are the
  sequences of (typically learnt) estimates of the values of the states
  visited along a batch of trajectories as observed by the agent given the
  sequences of one or more actions sampled from the `policy_logits`.

  The sequences in the tensors should be aligned such that an agent in a state
  with value `V` that takes an action `a` transitions into another state
  with value `V'`, receiving reward `r` and pcontinue `p`. Then `V`, `a`, `r`
  and `p` are all at the same index `i` in the corresponding tensors. `V'` is
  at index `i+1`, or in the `bootstrap_value` tensor if `i == T`.

  This function accepts a nested array of `policy_logits` and `actions` in order
  to allow for multidimensional discrete action spaces. In this case, the loss
  is given by `sum_i(loss(p_i, a_i))` where `p_i` are members of the
  `policy_logits` nest, and `a_i` are members of the `actions` nest.
  We assume that a single baseline is used across all action dimensions for
  each timestep.

  Args:
    policy_logits: A (possibly nested structure of) 3-D Tensor(s) with shape
        `[T, B, num_actions]` and possibly different dimension `num_actions`.
    baseline_values: 2-D Tensor containing an estimate of state values `[T, B]`.
    actions: A (possibly nested structure of) 2-D Tensor(s) with shape
        `[T, B]` and integer type.
    rewards: 2-D Tensor with shape `[T, B]`.
    pcontinues: 2-D Tensor with shape `[T, B]`.
    bootstrap_value: 1-D Tensor with shape `[B]`.
    lambda_: an optional scalar or 2-D Tensor with shape `[T, B]` for
        Generalised Advantage Estimation as per
        https://arxiv.org/abs/1506.02438.
    entropy_cost: optional scalar cost that pushes the policy to have high
        entropy, larger values cause higher entropies.
    baseline_cost: scalar cost that scales the derivatives of the baseline
        relative to the policy gradient.
    normalise_entropy: if True, the entropy loss is normalised to the range
        `[-1, 0]` by dividing by the log number of actions. This makes it more
        invariant to the size of the action space. Default is False.
    name: Customises the name_scope for this op.

  Returns:
    A namedtuple with fields:

    * `loss`: a tensor containing the total loss, shape `[B]`.
    * `extra`: a namedtuple with fields:
        * `entropy`: total loss per sequence, shape `[B]`.
        * `entropy_loss`: scaled entropy loss per sequence, shape `[B]`.
        * `baseline_loss`: scaled baseline loss per sequence, shape `[B]`.
        * `policy_gradient_loss`: policy gradient loss per sequence,
            shape `[B]`.
        * `advantages`: advantange estimates per timestep, shape `[T, B]`.
        * `discounted_returns`: discounted returns per timestep,
            shape `[T, B]`.
  """
    scoped_values = (
        nest.flatten(policy_logits)
        + nest.flatten(actions)
        + [baseline_values, rewards, pcontinues, bootstrap_value]
    )
    with tf.compat.v1.name_scope(name, values=scoped_values):
        # Loss for the baseline, summed over the time dimension.
        baseline_loss_td, td_lambda = value_ops.td_lambda(
            baseline_values, rewards, pcontinues, bootstrap_value, lambda_
        )

        # The TD error provides an estimate of the advantages of the actions.
        advantages = td_lambda.temporal_differences
        baseline_loss = tf.multiply(
            tf.convert_to_tensor(value=baseline_cost, dtype=tf.float32),
            baseline_loss_td,
            name="baseline_loss",
        )

        # Loss for the policy. Doesn't push additional gradients through
        # the advantages.
        policy_gradient_loss = discrete_policy_gradient_loss(
            policy_logits, actions, advantages, name="policy_gradient_loss"
        )

        total_loss = tf.add(policy_gradient_loss, baseline_loss, name="total_loss")

        if entropy_cost is not None:
            entropy_loss_op, policy_entropy = discrete_policy_entropy_loss(
                policy_logits, normalise=normalise_entropy
            )  # [T,B].
            entropy = tf.reduce_sum(
                input_tensor=policy_entropy.entropy, axis=0, name="entropy"
            )  # [B].
            entropy_loss = tf.multiply(
                tf.convert_to_tensor(value=entropy_cost, dtype=tf.float32),
                tf.reduce_sum(input_tensor=entropy_loss_op, axis=0),
                name="scaled_entropy_loss",
            )  # [B].
            total_loss = tf.add(
                total_loss, entropy_loss, name="total_loss_with_entropy"
            )
        else:
            entropy = None
            entropy_loss = None

        extra = SequenceAdvantageActorCriticExtra(
            entropy=entropy,
            entropy_loss=entropy_loss,
            baseline_loss=baseline_loss,
            policy_gradient_loss=policy_gradient_loss,
            advantages=advantages,
            discounted_returns=td_lambda.discounted_returns,
        )

        return base_ops.LossOutput(total_loss, extra)


def discrete_policy_gradient(
    policy_logits, actions, action_values, name="discrete_policy_gradient"
):
    """Computes a batch of discrete-action policy gradient losses.

  See notes by Silver et al here:
  http://www0.cs.ucl.ac.uk/staff/D.Silver/web/Teaching_files/pg.pdf

  From slide 41, denoting by `policy` the probability distribution with
  log-probabilities `policy_logit`:
  ```
  *   `action` should have been sampled according to `policy`.
  *   `action_value` can be any estimate of `Q^{policy}(s, a)`, potentially
      minus a baseline that doesn't depend on the action. This admits
      many possible algorithms:
      * `v_t` (Monte-Carlo return for time t) : REINFORCE
      * `Q^w(s, a)` : Q Actor-Critic
      * `v_t - V(s)` : Monte-Carlo Advantage Actor-Critic
      * `A^{GAE(gamma, lambda)}` : Generalized Avantage Actor Critic
      * + many more.
  ```

  Gradients for this op are only defined with respect to the `policy_logits`,
  not `actions` or `action_values`.

  This op supports multiple batch dimensions. The first N >= 1 dimensions of
  each input/output tensor index into independent values. All tensors must
  having matching sizes for each batch dimension.

  Args:
    policy_logits: (N+1)-D Tensor of shape
        `[batch_size_1, ..., batch_size_N, num_actions]` containing uncentered
        log-probabilities.
    actions: N-D Tensor of shape `[batch_size_1, ..., batch_size_N]` and integer
        type, containing indices for the selected actions.
    action_values: N-D Tensor of shape `[batch_size_1, ..., batch_size_N]`
        containing an estimate of the value of the selected `actions`.
    name: Customises the name_scope for this op.

  Returns:
    loss: N-D Tensor of shape `[batch_size_1, ..., batch_size_N]` containing the
        loss. Differentiable w.r.t `policy_logits` only.

  Raises:
    ValueError: If the batch dimensions of `policy_logits` and `action_values`
        do not match.
  """
    with tf.compat.v1.name_scope(name, values=[policy_logits, actions, action_values]):
        cross_entropy = tf.nn.sparse_softmax_cross_entropy_with_logits(
            labels=actions, logits=policy_logits
        )
        action_values = tf.stop_gradient(action_values)
        # Prevent accidental broadcasting if possible at construction time.
        action_values.get_shape().assert_is_compatible_with(cross_entropy.get_shape())
        return tf.multiply(cross_entropy, action_values)


def discrete_policy_gradient_loss(
    policy_logits, actions, action_values, name="discrete_policy_gradient_loss"
):
    """Computes discrete policy gradient losses for a batch of trajectories.

  This wraps `discrete_policy_gradient` to accept a possibly nested array of
  `policy_logits` and `actions` in order to allow for multiple discrete actions.
  It also sums up losses along the time dimension, and is more restrictive about
  shapes, assuming a [T, B] layout.

  Args:
    policy_logits: A (possibly nested structure of) Tensor(s) of shape
        `[T, B, num_actions]` containing uncentered log-probabilities.
    actions: A (possibly nested structure of) Tensor(s) of shape
        `[T, B]` and integer type, containing indices for the selected actions.
    action_values: Tensor of shape `[T, B]`
        containing an estimate of the value of the selected `actions`, see
        `discrete_policy_gradient`.
    name: Customises the name_scope for this op.

  Returns:
    loss: Tensor of shape `[B]` containing the total loss for each sequence
    in the batch. Differentiable w.r.t `policy_logits` only.
  """
    policy_logits = nest.flatten(policy_logits)
    actions = nest.flatten(actions)

    # Check happens after flatten so that we can be more flexible on
    # nest structures. This is equivalent to asserting that
    # `len(policy_logits) == len(actions)`, which is sufficient for what we're
    # doing here. In particular, it means that we can allow one argument to be
    # a tensor, while the other one to be a single-element tensor iterable.
    nest.assert_same_structure(policy_logits, actions)
    for scalar_policy_logits in policy_logits:
        scalar_policy_logits.get_shape().assert_has_rank(3)
    for scalar_actions in actions:
        scalar_actions.get_shape().assert_has_rank(2)

    scoped_values = policy_logits + actions + [action_values]
    with tf.compat.v1.name_scope(name, values=scoped_values):
        # Loss for the policy gradient. Doesn't push additional gradients through
        # the action_values.
        policy_gradient_loss_sequence = tf.add_n(
            [
                discrete_policy_gradient(
                    scalar_policy_logits, scalar_actions, action_values
                )
                for scalar_policy_logits, scalar_actions in zip(policy_logits, actions)
            ]
        )

        return tf.reduce_sum(
            input_tensor=policy_gradient_loss_sequence,
            axis=[0],
            name="policy_gradient_loss",
        )




from typing import Any, Dict, List, Optional, Type

from numpy.random import randint

from ..env import MathyEnvProblem
from ..state import MathyEnvState
from ..core.rule import BaseRule
from ..rules import (
    CommutativeSwapRule,
    ConstantsSimplifyRule,
    DistributiveFactorOutRule,
    AssociativeSwapRule,
)
from ..types import MathyEnvDifficulty, MathyEnvProblemArgs
from .poly_simplify import PolySimplify
from ..problems import gen_move_around_blockers_one, gen_move_around_blockers_two, rand_bool


class PolySimplifyBlockers(PolySimplify):
    """A Mathy environment for polynomial problems that have a variable
    string of mismatched terms separating two like terms.

    The goal is to:
      1. Commute the like terms so they become siblings
      2. Combine the sibling like terms
    """

    def get_env_namespace(self) -> str:
        return "mathy.polynomials.commute_then_simplify"

    def problem_fn(self, params: MathyEnvProblemArgs) -> MathyEnvProblem:
        hard_block = rand_bool()
        powers_probability = 0.5
        if params.difficulty == MathyEnvDifficulty.easy:
            powers_probability = 0.1
            blockers = randint(1, 3)
            hard_blockers = 1
        elif params.difficulty == MathyEnvDifficulty.normal:
            blockers = randint(2, 5)
            hard_blockers = randint(2, 3)
        elif params.difficulty == MathyEnvDifficulty.hard:
            blockers = randint(3, 7)
            hard_blockers = randint(2, 4)
        else:
            raise ValueError(f"Unknown difficulty: {params.difficulty}")
        if hard_block:
            text, complexity = gen_move_around_blockers_two(
                hard_blockers, powers_probability=powers_probability
            )
        else:
            text, complexity = gen_move_around_blockers_one(blockers, powers_probability)
        return MathyEnvProblem(text, complexity, self.get_env_namespace())



from typing import List, Optional

from pydantic import BaseModel
from .. import about


class BaseConfig(BaseModel):
    class Config:
        extra = "allow"

    name: str = "unnamed_model"
    # The version of the model. You probably don't want to change this.
    version: str = "0.0.1"
    description: str = "Mathy.ai trained model"
    license: str = "CC BY-SA 3.0"
    author: str = about.__author__
    email: str = about.__email__
    url: str = about.__uri__
    mathy_version: str = f">={about.__version__},<1.0.0"

    # One of "batch" or "layer"
    normalization_style = "layer"
    # The number of timesteps use when making predictions. This includes the current timestep and
    # (n - 1) previous timesteps
    prediction_window_size: int = 16

    # Dropout to apply to LSTMs
    dropout: float = 0.2
    units: int = 64
    embedding_units: int = 128
    lstm_units: int = 128
    topics: List[str] = ["poly"]
    difficulty: Optional[str] = None
    model_dir: str = "/tmp/a3c-training/"
    model_name: str = "model"
    init_model_from: Optional[str] = None
    verbose: bool = False
    # Initial learning rate that decays over time.
    lr_initial: float = 0.01
    lr_decay_steps: int = 100
    lr_decay_rate: float = 0.96
    lr_decay_staircase: bool = True
    max_eps: int = 15000
    # How often to write histograms to tensorboard (in training steps)
    summary_interval: int = 100
    gamma: float = 0.99
    # The number of worker agents to create.
    num_workers: int = 3
    # The lambda value for generalized lambda returns to calculate value loss
    # 0.0 = bootstrap values, 1.0 = discounted
    td_lambda: float = 0.2
    # Verbose setting to print out worker_0 training steps. Useful for trying
    # to find problems.
    print_training: bool = False
    # This is very verbose and prints every policy_value_model.call time
    print_model_call_times: bool = False
    # Print mode for output. "terminal" is the default, also supports "attention"
    # NOTE: attention is gone (like... the layer)
    print_mode: str = "terminal"



# coding=utf8
# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""TensorFlow ops for the Retrace algorithm.

These Retrace ops implement an action-value learning rule for minibatches of
multi-step sequences. Retrace supports off-policy learning and has
well-analysed theoretical guarantees.

The ops in this file support only discrete action spaces.  Actions must be
indices in the range `[0, K)`, where `K` is the number of distinct actions
available to the agent.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

# Dependency imports
from six.moves import zip
import tensorflow as tf
from . import base_ops
from . import indexing_ops
from . import sequence_ops

RetraceCoreExtra = collections.namedtuple(
    "retrace_core_extra", ["retrace_weights", "target"]
)


def retrace(
    lambda_,
    qs,
    targnet_qs,
    actions,
    rewards,
    pcontinues,
    target_policy_probs,
    behaviour_policy_probs,
    stop_targnet_gradients=True,
    name=None,
):
    """Retrace algorithm loss calculation op.

  Given a minibatch of temporally-contiguous sequences of Q values, policy
  probabilities, and various other typical RL algorithm inputs, this
  Op creates a subgraph that computes a loss according to the
  Retrace multi-step off-policy value learning algorithm. This Op supports the
  use of target networks, but does not require them.

  For more details of Retrace, refer to
  [the arXiv paper](http://arxiv.org/abs/1606.02647).

  In argument descriptions, `T` counts the number of transitions over which
  the Retrace loss is computed, and `B` is the minibatch size. Note that all
  tensor arguments list a first-dimension (time dimension) size of T+1;
  this is because in order to compute the loss over T timesteps, the
  algorithm must be aware of the values of many of its inputs at timesteps
  before and after each transition.

  All tensor arguments are indexed first by transition, with specific
  details of this indexing in the argument descriptions.

  Args:
    lambda_: Positive scalar value or 0-D `Tensor` controlling the degree to
      which future timesteps contribute to the loss computed at each
      transition.
    qs: 3-D tensor holding per-action Q-values for the states encountered
      just before taking the transitions that correspond to each major index.
      Since these values are the predicted values we wish to update (in other
      words, the values we intend to change as we learn), in a target network
      setting, these nearly always come from the "non-target" network, which
      we usually call the "learning network".
      Shape is `[(T+1), B, num_actions]`.
    targnet_qs: Like `qs`, but in the target network setting, these values
      should be computed by the target network. We use these values to
      compute multi-step error values for timesteps that follow the first
      timesteps in each sequence and sequence fragment we consider.
      Shape is `[(T+1), B, num_actions]`.
    actions: 2-D tensor holding the indices of actions executed during the
      transition that corresponds to each major index.
      Shape is `[(T+1), B]`.
    rewards: 2-D tensor holding rewards received during the transition
      that corresponds to each major index.
      Shape is `[(T+1), B]`.
    pcontinues: 2-D tensor holding pcontinue values received during the
      transition that corresponds to each major index.
      Shape is `[(T+1), B]`.
    target_policy_probs: 3-D tensor holding per-action policy probabilities
      for the states encountered just before taking the transitions that
      correspond to each major index, according to the target policy (i.e.
      the policy we wish to learn). These probabilities usually derive from
      the learning net.
      Shape is `[(T+1), B, num_actions]`.
    behaviour_policy_probs: 2-D tensor holding the *behaviour* policy's
      probabilities of having taken actions `action` during the transitions
      that correspond to each major index. These probabilities derive from
      whatever policy you used to generate the data.
      Shape is `[(T+1), B]`.
    stop_targnet_gradients: `bool` that enables a sensible default way of
      handling gradients through the Retrace op (essentially, gradients
      are not permitted to involve the `targnet_qs` inputs). Can be disabled
      if you require a different arrangement, but you'll probably want to
      block some gradients somewhere.
    name: name to prefix ops created by this function.

  Returns:
    A namedtuple with fields:

    * `loss`: Tensor containing the batch of losses, shape `[B]`.
    * `extra`: None
  """
    all_args = [
        lambda_,
        qs,
        targnet_qs,
        actions,
        rewards,
        pcontinues,
        target_policy_probs,
        behaviour_policy_probs,
    ]
    with tf.compat.v1.name_scope(name, "Retrace", values=all_args):
        # Mainly to simplify testing:
        (
            lambda_,
            qs,
            targnet_qs,
            actions,
            rewards,
            pcontinues,
            target_policy_probs,
            behaviour_policy_probs,
        ) = (tf.convert_to_tensor(value=arg) for arg in all_args)

        # Require correct tensor ranks---as long as we have shape information
        # available to check. If there isn't any, we print a warning.
        def check_rank(tensors, ranks):
            for i, (tensor, rank) in enumerate(zip(tensors, ranks)):
                if tensor.get_shape():
                    base_ops.assert_rank_and_shape_compatibility([tensor], rank)
                else:
                    tf.compat.v1.logging.error(
                        'Tensor "%s", which was offered as Retrace parameter %d, has '
                        "no rank at construction time, so Retrace can't verify that "
                        "it has the necessary rank of %d",
                        tensor.name,
                        i + 1,
                        rank,
                    )

        check_rank(
            [
                lambda_,
                qs,
                targnet_qs,
                actions,
                rewards,
                pcontinues,
                target_policy_probs,
                behaviour_policy_probs,
            ],
            [0, 3, 3, 2, 2, 2, 3, 2],
        )

        # Deduce the shapes of the arguments we'll create for retrace_core.
        qs_shape = tf.shape(input=qs)
        timesteps = qs_shape[0]  # Batch size is qs_shape[1].

        # Deduce the time indices for the arguments we'll create for retrace_core.
        timestep_indices_tm1 = tf.range(0, timesteps - 1)
        timestep_indices_t = tf.range(1, timesteps)

        # Construct arguments for retrace_core and call.
        q_tm1 = tf.gather(qs, timestep_indices_tm1)
        a_tm1 = tf.gather(actions, timestep_indices_tm1)

        r_t = tf.gather(rewards, timestep_indices_tm1)
        pcont_t = tf.gather(pcontinues, timestep_indices_tm1)

        target_policy_t = tf.gather(target_policy_probs, timestep_indices_t)
        behaviour_policy_t = tf.gather(behaviour_policy_probs, timestep_indices_t)
        targnet_q_t = tf.gather(targnet_qs, timestep_indices_t)
        a_t = tf.gather(actions, timestep_indices_t)

        core = retrace_core(
            lambda_,
            q_tm1,
            a_tm1,
            r_t,
            pcont_t,
            target_policy_t,
            behaviour_policy_t,
            targnet_q_t,
            a_t,
            stop_targnet_gradients,
        )

        return base_ops.LossOutput(core.loss, None)


def _general_off_policy_corrected_multistep_target(
    r_t, pcont_t, target_policy_t, c_t, q_t, a_t, back_prop=False, name=None
):
    """Evaluates targets for various off-policy value correction based algorithms.

  `target_policy_t` is the policy that this function aims to evaluate. New
  action-value estimates (target values `T`) must be expressible in this
  recurrent form:
  ```none
  T(x_{t-1}, a_{t-1}) = r_t + γ[ 𝔼_π Q(x_t, .) - c_t Q(x_t, a_t) +
                                                 c_t T(x_t, a_t) ]
  ```
  `T(x_t, a_t)` is an estimate of expected discounted future returns based
  on the current Q value estimates `Q(x_t, a_t)` and rewards `r_t`. The
  evaluated target values can be used as supervised targets for learning the Q
  function itself or as returns for various policy gradient algorithms.
  `Q==T` if convergence is reached. As the formula is recurrent, it will
  evaluate multistep returns for non-zero importance weights `c_t`.

  In the usual moving and target network setup `q_t` should be calculated by
  the target network while the `target_policy_t` may be evaluated by either of
  the networks. If `target_policy_t` is evaluated by the current moving network
  the algorithm implemented will have a similar flavour as double DQN.

  Depending on the choice of c_t, the algorithm can implement:
  ```none
  Importance Sampling             c_t = π(x_t, a_t) / μ(x_t, a_t),
  Harutyunyan's et al. Q(lambda)  c_t = λ,
  Precup's et al. Tree-Backup     c_t = π(x_t, a_t),
  Munos' et al. Retrace           c_t = λ min(1, π(x_t, a_t) / μ(x_t, a_t)).
  ```
  Please refer to page 3 for more details:
  https://arxiv.org/pdf/1606.02647v1.pdf

  Args:
    r_t: 2-D tensor holding rewards received during the transition
      that corresponds to each major index.
      Shape is `[T, B]`.
    pcont_t: 2-D tensor holding pcontinue values received during the
      transition that corresponds to each major index.
      Shape is `[T, B]`.
    target_policy_t:  3-D tensor holding per-action policy probabilities for
      the states encountered just AFTER the transitions that correspond to
      each major index, according to the target policy (i.e. the policy we
      wish to learn). These usually derive from the learning net.
      Shape is `[T, B, num_actions]`.
    c_t: 2-D tensor holding importance weights; see discussion above.
      Shape is `[T, B]`.
    q_t: 3-D tensor holding per-action Q-values for the states
      encountered just AFTER taking the transitions that correspond to each
      major index. Shape is `[T, B, num_actions]`.
    a_t: 2-D tensor holding the indices of actions executed during the
      transition AFTER the transition that corresponds to each major index.
      Shape is `[T, B]`.
    back_prop: whether to backpropagate gradients through time.
    name: name of the op.

  Returns:
    Tensor of shape `[T, B, num_actions]` containing Q values.
  """
    # Formula (4) in https://arxiv.org/pdf/1606.02647v1.pdf can be expressed
    # in a recursive form where T is a new target value:
    # T(x_{t-1}, a_{t-1}) = r_t + γ[ 𝔼_π Q(x_t, .) - c_t Q(x_t, a_t) +
    #                                                c_t T(x_t, a_t) ]
    # This recurrent form allows us to express Retrace by using
    # `scan_discounted_sum`.
    # Define:
    #   T_tm1   = T(x_{t-1}, a_{t-1})
    #   T_t     = T(x_t, a_t)
    #   exp_q_t = 𝔼_π Q(x_{t+1},.)
    #   qa_t    = Q(x_t, a_t)
    # Hence:
    #   T_tm1   = (r_t + γ * exp_q_t - c_t * qa_t) + γ * c_t * T_t
    # Define:
    #   current = r_t + γ * (exp_q_t - c_t * qa_t)
    # Thus:
    #   T_tm1 = scan_discounted_sum(current, γ * c_t, reverse=True)
    args = [r_t, pcont_t, target_policy_t, c_t, q_t, a_t]
    with tf.compat.v1.name_scope(
        name, "general_returns_based_off_policy_target", values=args
    ):
        exp_q_t = tf.reduce_sum(input_tensor=target_policy_t * q_t, axis=2)
        qa_t = indexing_ops.batched_index(q_t, a_t)
        current = r_t + pcont_t * (exp_q_t - c_t * qa_t)
        initial_value = qa_t[-1]
        return sequence_ops.scan_discounted_sum(
            current, pcont_t * c_t, initial_value, reverse=True, back_prop=back_prop
        )


def _retrace_weights(pi_probs, mu_probs):
    """Evaluates importance weights for the Retrace algorithm.

  Args:
    pi_probs: taken action probabilities according to target policy.
      Shape is `[T, B]`.
    mu_probs: taken action probabilities according to behaviour policy.
      Shape is `[T, B]`.

  Returns:
    Tensor of shape `[T, B]` containing importance weights.
  """
    # tf.minimum seems to handle potential NaNs when pi_probs[i]=mu_probs[i]=0
    return tf.minimum(1.0, pi_probs / mu_probs)


def retrace_core(
    lambda_,
    q_tm1,
    a_tm1,
    r_t,
    pcont_t,
    target_policy_t,
    behaviour_policy_t,
    targnet_q_t,
    a_t,
    stop_targnet_gradients=True,
    name=None,
):
    """Retrace algorithm core loss calculation op.

  Given a minibatch of temporally-contiguous sequences of Q values, policy
  probabilities, and various other typical RL algorithm inputs, this
  Op creates a subgraph that computes a loss according to the
  Retrace multi-step off-policy value learning algorithm. This Op supports the
  use of target networks, but does not require them.

  This function is the "core" Retrace op only because its arguments are less
  user-friendly and more implementation-convenient. For a more user-friendly
  operator, consider using `retrace`. For more details of Retrace, refer to
  [the arXiv paper](http://arxiv.org/abs/1606.02647).

  Construct the "core" retrace loss subgraph for a batch of sequences.

  Note that two pairs of arguments (one holding target network values; the
  other, actions) are temporally-offset versions of each other and will share
  many values in common (nb: a good setting for using `IndexedSlices`). *This
  op does not include any checks that these pairs of arguments are
  consistent*---that is, it does not ensure that temporally-offset
  arguments really do share the values they are supposed to share.

  In argument descriptions, `T` counts the number of transitions over which
  the Retrace loss is computed, and `B` is the minibatch size. All tensor
  arguments are indexed first by transition, with specific details of this
  indexing in the argument descriptions (pay close attention to "subscripts"
  in variable names).

  Args:
    lambda_: Positive scalar value or 0-D `Tensor` controlling the degree to
      which future timesteps contribute to the loss computed at each
      transition.
    q_tm1: 3-D tensor holding per-action Q-values for the states encountered
      just before taking the transitions that correspond to each major index.
      Since these values are the predicted values we wish to update (in other
      words, the values we intend to change as we learn), in a target network
      setting, these nearly always come from the "non-target" network, which
      we usually call the "learning network".
      Shape is `[T, B, num_actions]`.
    a_tm1: 2-D tensor holding the indices of actions executed during the
      transition that corresponds to each major index.
      Shape is `[T, B]`.
    r_t: 2-D tensor holding rewards received during the transition
      that corresponds to each major index.
      Shape is `[T, B]`.
    pcont_t: 2-D tensor holding pcontinue values received during the
      transition that corresponds to each major index.
      Shape is `[T, B]`.
    target_policy_t: 3-D tensor holding per-action policy probabilities for
      the states encountered just AFTER the transitions that correspond to
      each major index, according to the target policy (i.e. the policy we
      wish to learn). These usually derive from the learning net.
      Shape is `[T, B, num_actions]`.
    behaviour_policy_t: 2-D tensor holding the *behaviour* policy's
      probabilities of having taken action `a_t` at the states encountered
      just AFTER the transitions that correspond to each major index. Derived
      from whatever policy you used to generate the data. All values MUST be
      greater that 0. Shape is `[T, B]`.
    targnet_q_t: 3-D tensor holding per-action Q-values for the states
      encountered just AFTER taking the transitions that correspond to each
      major index. Since these values are used to calculate target values for
      the network, in a target in a target network setting, these should
      probably come from the target network.
      Shape is `[T, B, num_actions]`.
    a_t: 2-D tensor holding the indices of actions executed during the
      transition AFTER the transition that corresponds to each major index.
      Shape is `[T, B]`.
    stop_targnet_gradients: `bool` that enables a sensible default way of
      handling gradients through the Retrace op (essentially, gradients
      are not permitted to involve the `targnet_q_t` input).
      Can be disabled if you require a different arragement, but
      you'll probably want to block some gradients somewhere.
    name: name to prefix ops created by this function.

  Returns:
    A namedtuple with fields:

    * `loss`: Tensor containing the batch of losses, shape `[B]`.
    * `extra`: A namedtuple with fields:
        * `retrace_weights`: Tensor containing batch of retrace weights,
        shape `[T, B]`.
        * `target`: Tensor containing target action values, shape `[T, B]`.
  """
    all_args = [
        lambda_,
        q_tm1,
        a_tm1,
        r_t,
        pcont_t,
        target_policy_t,
        behaviour_policy_t,
        targnet_q_t,
        a_t,
    ]

    with tf.compat.v1.name_scope(name, "RetraceCore", all_args):
        (
            lambda_,
            q_tm1,
            a_tm1,
            r_t,
            pcont_t,
            target_policy_t,
            behaviour_policy_t,
            targnet_q_t,
            a_t,
        ) = (tf.convert_to_tensor(value=arg) for arg in all_args)

        # Evaluate importance weights.
        c_t = (
            _retrace_weights(
                indexing_ops.batched_index(target_policy_t, a_t), behaviour_policy_t
            )
            * lambda_
        )
        # Targets are evaluated by using only Q values from the target network.
        # This provides fixed regression targets until the next target network
        # update.
        target = _general_off_policy_corrected_multistep_target(
            r_t,
            pcont_t,
            target_policy_t,
            c_t,
            targnet_q_t,
            a_t,
            not stop_targnet_gradients,
        )

        if stop_targnet_gradients:
            target = tf.stop_gradient(target)
        # Regress Q values of the learning network towards the targets evaluated
        # by using the target network.
        qa_tm1 = indexing_ops.batched_index(q_tm1, a_tm1)
        delta = target - qa_tm1
        loss = 0.5 * tf.square(delta)

        return base_ops.LossOutput(
            loss, RetraceCoreExtra(retrace_weights=c_t, target=target)
        )




import datetime
import multiprocessing
from dataclasses import dataclass, field
from typing import Dict

from colr import color


@dataclass
class EpisodeLosses:
    """Store a set of losses keyed by a string that is used when printing them"""

    data: Dict[str, float] = field(default_factory=dict)

    def reset(self):
        self.data = dict()

    def increment(self, key: str, value: float) -> None:
        if not key in self.data:
            self.data[key] = 0.0
        self.data[key] += value

    def format_loss(self, key: str, value: float):
        return "{}: {:<8}".format(key, truncate(value))

    def __str__(self):
        out = ""
        for key, value in self.data.items():
            out += f"{self.format_loss(key, value)} "
        return out


def truncate(value):
    return float("%.3f" % (float(value)))


def record(
    episode: int,
    is_win: bool,
    episode_reward: float,
    worker_idx: int,
    global_ep_reward: float,
    result_queue: multiprocessing.Queue,
    losses: EpisodeLosses,
    num_steps: int,
    env_name: str,
):
    """Helper function to store score and print statistics.
  Arguments:
    episode: Current episode
    episode_reward: Reward accumulated over the current episode
    worker_idx: Which thread (worker)
    global_ep_reward: The moving average of the global reward
    result_queue: Queue storing the moving average of the scores
    total_loss: The total loss accumualted over the current episode
    num_steps: The number of steps the episode took to complete
  """

    now = datetime.datetime.now().strftime("%H:%M:%S")
    # Clamp to range -2, 2
    episode_reward = min(2.0, max(-2.0, episode_reward))

    global_ep_reward = global_ep_reward * 0.99 + episode_reward * 0.01

    fore = "green"
    if not is_win:
        fore = "red"
    elif episode_reward < 1.0:
        fore = "yellow"
    heading = "{:<8} {:<3} {:<8} {:<10}".format(
        now, f"w{worker_idx}", f"ep: {episode}", f"steps: {num_steps}"
    )
    rewards = "r_avg: {:<6} r_ep: {:<6}".format(
        truncate(global_ep_reward), truncate(episode_reward)
    )
    print(
        color(f"{heading} {rewards} {losses} [{env_name}]", fore=fore, style="bright")
    )
    result_queue.put(global_ep_reward)
    return global_ep_reward



# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Indexing ops.

These ops support indexing the 2D tensors representing batches of values
(shape: [B, dim]) or 3D tensors representing batches of sequences
of values (shape: [T, B, dim]. `T` is the length of the rollout, `B` is the
batch size, and `dim` the size of the dimension that must be indexed.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

# Dependency imports
import tensorflow as tf


def assert_compatible_shapes(value_shape, index_shape):
    """Check shapes of the indices and the tensor to be indexed.

  If all input shapes are known statically, obtain shapes of arguments and
  perform compatibility checks. Otherwise, print a warning. The only check
  we cannot perform statically (and do not attempt elsewhere) is making
  sure that each action index in actions is in [0, num_actions).

  Args:
    value_shape: static shape of the values.
    index_shape: static shape of the indices.
  """
    # note: rank-0 "[]" TensorShape is still True.
    if value_shape and index_shape:
        try:
            msg = (
                'Shapes of "values" and "indices" do not correspond to '
                "minibatch (2-D) or sequence-minibatch (3-D) indexing"
            )
            assert (value_shape.ndims, index_shape.ndims) in [(2, 1), (3, 2)], msg
            msg = (
                '"values" and "indices" have incompatible shapes of {} '
                "and {}, respectively"
            ).format(value_shape, index_shape)
            assert value_shape[:-1].is_compatible_with(index_shape), msg
        except AssertionError as e:
            raise ValueError(e)  # Convert AssertionError to ValueError.

    else:  # No shape information is known ahead of time.
        tf.compat.v1.logging.warning(
            'indexing function cannot get shapes for tensors "values" and '
            '"indices" at construction time, and so can\'t check that their '
            "shapes are valid or compatible. Incorrect indexing may occur at "
            "runtime without error!"
        )


def batched_index(values, indices):
    """Equivalent to `values[:, indices]`.

  Performs indexing on batches and sequence-batches by reducing over
  zero-masked values. Compared to indexing with `tf.gather` this approach is
  more general and TPU-friendly, but may be less efficient if `num_values`
  is large. It works with tensors whose shapes are unspecified or
  partially-specified, but this op will only do shape checking on shape
  information available at graph construction time. When complete shape
  information is absent, certain shape incompatibilities may not be detected at
  runtime! See `indexing_ops_test` for detailed examples.

  Args:
    values: tensor of shape `[B, num_values]` or `[T, B, num_values]`
    indices: tensor of shape `[B]` or `[T, B]` containing indices.

  Returns:
    Tensor of shape `[B]` or `[T, B]` containing values for the given indices.

  Raises: ValueError if values and indices have sizes that are known
    statically (i.e. during graph construction), and those sizes are not
    compatible (see shape descriptions in Args list above).
  """
    with tf.compat.v1.name_scope("batch_indexing", values=[values, indices]):
        values = tf.convert_to_tensor(value=values)
        indices = tf.convert_to_tensor(value=indices)
        assert_compatible_shapes(values.shape, indices.shape)

        one_hot_indices = tf.one_hot(
            indices, tf.shape(input=values)[-1], dtype=values.dtype
        )
        return tf.reduce_sum(input_tensor=values * one_hot_indices, axis=-1)



from ..mathy.teacher import Student, Teacher

topic_names = ["poly", "binomial", "poly-blockers", "complex"]
me = 0


def test_teacher_env_rotation_by_iteration_modulus():
    teacher = Teacher(topic_names, num_students=2)

    # Test env rotation
    for i, e in enumerate(topic_names):
        # rotation is modulus by iteration
        result = teacher.get_env(1, i)
        assert result == f"mathy-{e}-easy-v0"


def test_teacher_evaluation_window_win_loss_record():
    teacher = Teacher(topic_names, eval_window=10)
    # Report win/loss
    teacher.report_result(me, 1.0)
    teacher.report_result(me, -0.1)
    teacher.report_result(me, 1.0)

    student: Student = teacher.students[me]
    assert student.topics[student.topic].positives == 2
    assert student.topics[student.topic].negatives == 1


def test_teacher_evaluation_objects():
    teacher = Teacher(topic_names, eval_window=10)
    # Report win/loss
    teacher.report_result(me, 1.0)
    teacher.report_result(me, -0.1)
    teacher.report_result(me, 1.0)

    student: Student = teacher.students[me]
    assert student.topics[student.topic].positives == 2
    assert student.topics[student.topic].negatives == 1


def test_teacher_evaluation_window_change_difficulty():
    teacher = Teacher(
        topic_names, eval_window=10, win_threshold=0.7, lose_threshold=0.5
    )
    # Report win/loss
    student = teacher.students[me]

    assert student.topics[student.topic].difficulty == "easy"

    # 60% wins remains at the same difficulty
    for i in range(4):
        teacher.report_result(me, -1.0)
    for i in range(6):
        teacher.report_result(me, 1.0)

    assert student.topics[student.topic].difficulty == "easy"

    # >= 70% wins goes to higher difficulty
    for i in range(7):
        teacher.report_result(me, 1.0)
    for i in range(3):
        teacher.report_result(me, -1.0)

    assert student.topics[student.topic].difficulty == "normal"

    # <= 50% wins goes to easier difficulty
    for i in range(6):
        teacher.report_result(me, -1.0)
    for i in range(4):
        teacher.report_result(me, 1.0)

    assert student.topics[student.topic].difficulty == "easy"



from ..mathy.state import MathyEnvState
from ..mathy.env import MathyEnv
from ..mathy.envs.poly_simplify import PolySimplify
from ..mathy.util import is_terminal_transition
import random
import pytest


def test_mathy_env_init():
    env = MathyEnv()
    assert env is not None
    # Default env is abstract and cannot be directly used for problem solving
    with pytest.raises(NotImplementedError):
        env.get_initial_state()
    with pytest.raises(NotImplementedError):
        env.get_env_namespace()


def test_env_terminal_conditions():

    expectations = [
        ("70656 * (x^2 * z^6)", True),
        ("b * (44b^2)", False),
        ("z * (1274z^2)", False),
        ("4x^2", True),
        ("100y * x + 2", True),
        ("10y * 10x + 2", False),
        ("10y + 1000y * (y * z)", False),
        ("4 * (5y + 2)", False),
        ("2", True),
        ("4x * 2", False),
        ("4x * 2x", False),
        ("4x + 2x", False),
        ("4 + 2", False),
        ("3x + 2y + 7", True),
        ("3x^2 + 2x + 7", True),
        ("3x^2 + 2x^2 + 7", False),
    ]

    # Valid solutions but out of scope so they aren't counted as wins.
    #
    # This works because the problem sets exclude this type of > 2 term
    # polynomial expressions
    out_of_scope_valid = []

    env = PolySimplify()
    for text, is_win in expectations + out_of_scope_valid:
        env_state = MathyEnvState(problem=text)
        reward = env.get_state_transition(env_state)
        assert text == text and env.is_terminal_state(env_state) == bool(is_win)
        assert text == text and is_terminal_transition(reward) == bool(is_win)


def test_print_history():
    env = PolySimplify()
    env_state = MathyEnvState(problem="4x+2")
    for i in range(10):
        env_state = env_state.get_out_state(
            problem="2+4x", focus=i, moves_remaining=10 - i, action=i
        )
    assert env.print_history(env_state) is None


def test_env_finalize_state():
    env = PolySimplify()

    env_state = MathyEnvState(problem="4x + 2x").get_out_state(
        problem="1337", action=1, focus=-1, moves_remaining=0
    )
    with pytest.raises(ValueError):
        env.finalize_state(env_state)

    env_state = MathyEnvState(problem="4x + 2x").get_out_state(
        problem="4x + 2", action=1, focus=-1, moves_remaining=0
    )
    with pytest.raises(ValueError):
        env.finalize_state(env_state)

    env_state = MathyEnvState(problem="4x + 2x").get_out_state(
        problem="4x + 2y", action=1, focus=-1, moves_remaining=0
    )
    with pytest.raises(ValueError):
        env.finalize_state(env_state)



import pytest
import numpy as np
from ..mathy.agents.densenet import DenseNetBlock, DenseNetStack
import tensorflow as tf


def test_densenet_construction():
    layer: DenseNetStack = DenseNetStack()
    assert layer is not None
    model = tf.keras.Sequential([layer])
    model(np.zeros(shape=(128, 1)))


def test_densenet_errors():
    with pytest.raises(ValueError):
        DenseNetStack(normalization_style="invalid")


@pytest.mark.parametrize("norm_type", ("batch", "layer"))
def test_densenet_normalization_types(norm_type: str):
    model = tf.keras.Sequential(
        [DenseNetStack(units=24, num_layers=4, normalization_style=norm_type)]
    )
    y = model(np.zeros(shape=(10, 1)))


def test_densenet_no_activation():
    model = tf.keras.Sequential(
        [DenseNetStack(units=24, num_layers=4, activation=None)]
    )
    y = model(np.zeros(shape=(10, 1)))
    assert y.shape[0] == 10


def test_densenet_output_transform():
    layer: DenseNetStack = DenseNetStack(
        units=10,
        num_layers=4,
        activation=None,
        output_transform=tf.keras.layers.Dense(128),
    )
    assert layer is not None
    model = tf.keras.Sequential([layer])
    x = np.zeros(shape=(10, 1))
    y = model(x)
    assert y.shape[1] == 128


def test_densenet_share_weights():
    layer: DenseNetStack = DenseNetStack(share_weights=True)
    layer_two: DenseNetStack = DenseNetStack(share_weights=True)
    assert layer.get_weights() == layer_two.get_weights()
    model = tf.keras.Sequential([layer, layer_two])
    x = np.zeros(shape=(10, 1))
    y = model(x)




from enum import IntEnum
from typing import Any, Dict, List, NamedTuple, Optional

import numpy as np
import srsly

from .core.expressions import ConstantExpression, MathExpression, MathTypeKeys
from .core.parser import ExpressionParser
from .util import pad_array

PROBLEM_TYPE_HASH_BUCKETS = 128

NodeIntList = List[int]
NodeValuesFloatList = List[float]
NodeMaskIntList = List[int]
ProblemTypeIntList = List[int]


WindowNodeIntList = List[NodeIntList]
WindowNodeMaskIntList = List[NodeMaskIntList]
WindowNodeValuesFloatList = List[NodeValuesFloatList]
WindowProblemTypeIntList = List[ProblemTypeIntList]


# Input type for mathy models
MathyInputsType = List[Any]


class ObservationFeatureIndices(IntEnum):
    nodes = 0
    mask = 1
    values = 2
    type = 3


class MathyObservation(NamedTuple):
    """A featurized observation from an environment state."""

    nodes: NodeIntList
    mask: NodeMaskIntList
    values: NodeValuesFloatList
    type: ProblemTypeIntList


# fmt: off
MathyObservation.nodes.__doc__ = "tree node types in the current environment state shape=[n,]" # noqa
MathyObservation.mask.__doc__ = "0/1 mask where 0 indicates an invalid action shape=[n,]" # noqa
MathyObservation.values.__doc__ = "tree node value sequences, with non number indices set to 0.0 shape=[n,]" # noqa
MathyObservation.type.__doc__ = "two column hash of problem environment type shape=[2,]" # noqa
# fmt: on


class MathyWindowObservation(NamedTuple):
    """A featurized observation from an n-step sequence of environment states."""

    nodes: WindowNodeIntList
    mask: WindowNodeMaskIntList
    values: WindowNodeValuesFloatList
    type: WindowProblemTypeIntList

    def to_inputs(self, as_tf_tensor: bool = False) -> MathyInputsType:
        if as_tf_tensor:
            import tensorflow as tf

        def to_res(in_value):
            if as_tf_tensor is True:
                return tf.convert_to_tensor(in_value, dtype=tf.float32)
            return np.asarray(in_value, dtype="float32")

        result = [
            to_res(self.nodes),
            to_res(self.mask),
            to_res(self.values),
            to_res(self.type),
        ]
        for r in result:
            for s in r.shape:
                assert s is not None
        return result

    def to_input_shapes(self) -> List[Any]:
        import tensorflow as tf

        result = [
            np.asarray(self.nodes).shape,
            np.asarray(self.mask).shape,
            np.asarray(self.values).shape,
            np.asarray(self.type).shape,
        ]
        return result


# fmt: off
MathyWindowObservation.nodes.__doc__ = "n-step list of node sequences `shape=[n, max(len(s))]`" # noqa
MathyWindowObservation.mask.__doc__ = "n-step list of node sequence masks `shape=[n, max(len(s))]`" # noqa
MathyWindowObservation.values.__doc__ = "n-step list of value sequences, with non number indices set to 0.0 `shape=[n, max(len(s))]`" # noqa
MathyWindowObservation.type.__doc__ = "n-step problem type hash `shape=[n, 2]`" # noqa
# fmt: on


class MathyEnvStateStep(NamedTuple):
    """Capture summarized environment state for a previous timestep so the
    agent can use context from its history when making new predictions."""

    raw: str
    focus: int
    action: int


# fmt: off
MathyEnvStateStep.raw.__doc__ = "the input text at the timestep" # noqa
MathyEnvStateStep.focus.__doc__ = "the index of the node that is acted on" # noqa
MathyEnvStateStep.action.__doc__ = "the action taken" # noqa
# fmt: on


_problem_hash_cache: Optional[Dict[str, List[int]]] = None


def observations_to_window(
    observations: List[MathyObservation], total_length: int = None
) -> MathyWindowObservation:
    """Combine a sequence of observations into an observation window"""
    output = MathyWindowObservation(nodes=[], mask=[], values=[], type=[])
    sequence_lengths: List[int] = []
    max_mask_length: int = 0
    for i in range(len(observations)):
        obs = observations[i]
        sequence_lengths.append(len(obs.nodes))
        sequence_lengths.append(len(obs.values))
        max_mask_length = max(max_mask_length, len(obs.mask))
    max_length: int = max(sequence_lengths)
    if total_length is not None:
        max_length = total_length

    for obs in observations:
        output.nodes.append(pad_array(obs.nodes, max_length, MathTypeKeys["empty"]))
        output.mask.append(pad_array(obs.mask, max_mask_length, 0))
        output.values.append(pad_array(obs.values, max_length, 0.0))
        output.type.append(pad_array([], max_length, obs.type))
    return output


class MathyEnvState(object):
    """Class for holding environment state and extracting features
    to be passed to the policy/value neural network.

    Mutating operations all return a copy of the environment adapter
    with its own state.

    This allocation strategy requires more memory but removes a class
    of potential issues around unintentional sharing of data and mutation
    by two different sources.
    """

    agent: "MathyAgentState"
    max_moves: int
    num_rules: int

    def __init__(
        self,
        state: Optional["MathyEnvState"] = None,
        problem: str = None,
        max_moves: int = 10,
        num_rules: int = 0,
        problem_type: str = "mathy.unknown",
    ):
        self.max_moves = max_moves
        self.num_rules = num_rules
        if problem is not None:
            self.agent = MathyAgentState(max_moves, problem, problem_type)
        elif state is not None:
            self.num_rules = state.num_rules
            self.max_moves = state.max_moves
            self.agent = MathyAgentState.copy(state.agent)

    @classmethod
    def copy(cls, from_state):
        return MathyEnvState(state=from_state)

    def clone(self):
        return MathyEnvState(state=self)

    def get_out_state(
        self, problem: str, focus: int, action: int, moves_remaining: int
    ) -> "MathyEnvState":
        """Get the next environment state based on the current one with updated
        history and agent information based on an action being taken."""
        out_state = MathyEnvState.copy(self)
        agent = out_state.agent
        agent.history.append(MathyEnvStateStep(problem, focus, action))
        agent.problem = problem
        agent.action = action
        agent.moves_remaining = moves_remaining
        return out_state

    def get_problem_hash(self) -> ProblemTypeIntList:
        """Return a two element array with hashed values for the current environment
        namespace string.

        # Example

        - `mycorp.envs.solve_impossible_problems` -> `[12375561, -2838517]`

        """
        # NOTE: import delayed to enable MP
        import tensorflow as tf

        global _problem_hash_cache
        if _problem_hash_cache is None:
            _problem_hash_cache = {}
        if self.agent.problem_type not in _problem_hash_cache:
            problem_hash_one = tf.strings.to_hash_bucket_strong(
                self.agent.problem_type, PROBLEM_TYPE_HASH_BUCKETS, [1337, 61059873]
            )
            problem_hash_two = tf.strings.to_hash_bucket_strong(
                self.agent.problem_type, PROBLEM_TYPE_HASH_BUCKETS, [7, -242315]
            )
            _problem_hash_cache[self.agent.problem_type] = [
                int(problem_hash_one.numpy()),
                int(problem_hash_two.numpy()),
            ]
        return _problem_hash_cache[self.agent.problem_type]

    def to_start_observation(self) -> MathyObservation:
        """Generate an episode start MathyObservation"""
        num_actions = 1 * self.num_rules
        hash = self.get_problem_hash()
        mask = [0] * num_actions
        values = [0.0]
        return MathyObservation(
            nodes=[MathTypeKeys["empty"]], mask=mask, values=values, type=hash,
        )

    def to_observation(
        self,
        move_mask: Optional[NodeMaskIntList] = None,
        hash_type: Optional[ProblemTypeIntList] = None,
        parser: Optional[ExpressionParser] = None,
    ) -> MathyObservation:
        """Convert a state into an observation"""
        if parser is None:
            parser = ExpressionParser()
        if hash_type is None:
            hash_type = self.get_problem_hash()
        expression = parser.parse(self.agent.problem)
        nodes: List[MathExpression] = expression.to_list()
        vectors: NodeIntList = []
        values: NodeValuesFloatList = []
        if move_mask is None:
            move_mask = np.zeros(len(nodes))
        for node in nodes:
            vectors.append(node.type_id)
            if isinstance(node, ConstantExpression):
                values.append(float(node.value))
            else:
                values.append(0.0)

        return MathyObservation(
            nodes=vectors, mask=move_mask, values=values, type=hash_type
        )

    @classmethod
    def from_string(cls, input_string: str) -> "MathyEnvState":
        """Convert a string representation of state into a state object"""
        sep = "@"
        history_sep = ","
        inputs = input_string.split(sep)
        state = MathyEnvState()
        state.max_moves = int(inputs[0])
        state.num_rules = int(inputs[1])
        state.agent = MathyAgentState(
            moves_remaining=int(inputs[2]),
            problem=str(inputs[3]),
            problem_type=str(inputs[4]),
            reward=float(inputs[5]),
            history=[],
        )
        history = inputs[6:]
        for step in history:
            raw, focus, action = step.split(history_sep)
            state.agent.history.append(MathyEnvStateStep(raw, int(focus), int(action)))
        return state

    def to_string(self) -> str:
        """Convert a state object into a string representation"""
        sep = "@"
        out = [
            str(self.max_moves),
            str(self.num_rules),
            str(self.agent.moves_remaining),
            str(self.agent.problem),
            str(self.agent.problem_type),
            str(self.agent.reward),
        ]
        for step in self.agent.history:
            out.append(",".join([str(step.raw), str(step.focus), str(step.action)]))
        return sep.join(out)

    @classmethod
    def from_np(cls, input_bytes: np.ndarray) -> "MathyEnvState":
        """Convert a numpy object into a state object"""
        input_string = "".join([chr(o) for o in input_bytes.tolist()])
        state = cls.from_string(input_string)
        return state

    def to_np(self) -> np.ndarray:
        """Convert a state object into a numpy representation"""
        return np.array([ord(c) for c in self.to_string()])


class MathyAgentState:
    """The state related to an agent for a given environment state"""

    moves_remaining: int
    problem: str
    problem_type: str
    reward: float
    history: List[MathyEnvStateStep]

    def __init__(
        self, moves_remaining, problem, problem_type, reward=0.0, history=None,
    ):
        self.moves_remaining = moves_remaining
        self.problem = problem
        self.reward = reward
        self.problem_type = problem_type
        self.history = (
            history[:] if history is not None else [MathyEnvStateStep(problem, -1, -1)]
        )

    @classmethod
    def copy(cls, from_state: "MathyAgentState"):
        return MathyAgentState(
            moves_remaining=from_state.moves_remaining,
            problem=from_state.problem,
            reward=from_state.reward,
            problem_type=from_state.problem_type,
            history=from_state.history,
        )



import math
from typing import Dict, List, Optional, Tuple, Type, Union, cast, TypeVar

import numpy as np
from colr import color

from .tree import STOP, BinaryTreeNode

OOO_FUNCTION = 4
OOO_PARENS = 3
OOO_EXPONENT = 2
OOO_MULTDIV = 1
OOO_ADDSUB = 0
OOO_INVALID = -1


MathTypeKeys = {
    "empty": 0,
    "negate": 1,
    "equal": 2,
    "add": 3,
    "subtract": 4,
    "multiply": 5,
    "divide": 6,
    "power": 7,
    "term_root": 8,
    "term_connector": 9,
    "constant": 10,
    "sgn": 11,
    "abs": 12,
    # NOTE: reserved 13-25 for future expression types (such as functions)
    "variable": 26,
    "variable_a": 27,
    "variable_b": 28,
    "variable_c": 29,
    "variable_d": 30,
    "variable_e": 31,
    "variable_f": 32,
    "variable_g": 33,
    "variable_h": 34,
    "variable_i": 35,
    "variable_j": 36,
    "variable_k": 37,
    "variable_l": 38,
    "variable_m": 39,
    "variable_n": 40,
    "variable_o": 41,
    "variable_p": 42,
    "variable_q": 43,
    "variable_r": 44,
    "variable_s": 45,
    "variable_t": 46,
    "variable_u": 47,
    "variable_v": 48,
    "variable_w": 49,
    "variable_x": 50,
    "variable_y": 51,
    "variable_z": 52,
}


# The maximum value in type keys (for one-hot encoding)
MathTypeKeysMax = max(MathTypeKeys.values()) + 1

NodeType = TypeVar("NodeType", bound="MathExpression")


class MathExpression(BinaryTreeNode):
    """Math tree node with helpers for manipulating expressions.
    
    `mathy:x+y=z`
    """

    left: Optional["MathExpression"]
    right: Optional["MathExpression"]
    parent: Optional["MathExpression"]

    @property
    def raw(self) -> str:
        """raw text representation of the expression."""
        return str(self)

    @property
    def type_id(self) -> int:
        raise NotImplementedError("must be implemented in subclass")

    @property
    def terminal_text(self) -> str:
        """Text output of this node that includes terminal color codes that
        highlight which nodes have been changed in this tree as a result of
        a transformation."""

        def visit_fn(node, depth, data):
            node._rendering_change = data

        self.visit_inorder(visit_fn, data=True)
        result = str(self)
        self.visit_inorder(visit_fn, data=False)
        return result

    @property
    def color(self) -> str:
        """Color to use for this node when rendering it as changed with `.terminal_text`"""
        return "green"

    _rendering_change: bool
    _changed: bool
    classes: List[str]
    cloned_node: Optional["MathExpression"]
    cloned_target: str

    def __init__(
        self,
        id: str = None,
        left: "MathExpression" = None,
        right: "MathExpression" = None,
        parent: "MathExpression" = None,
    ):
        super().__init__(left, right, parent, id)
        self._rendering_change = False
        self._changed = False
        self.classes = [self.id]
        self.cloned_node = None
        self.cloned_target = ""

    def evaluate(
        self, context: Dict[str, Union[float, int]] = None
    ) -> Union[float, int]:
        """Evaluate the expression, resolving all variables to constant values"""
        raise NotImplementedError("must be implemented in subclass")

    def set_changed(self) -> None:
        """Mark this node as having been changed by the application of a Rule"""
        self._changed = True

    def all_changed(self) -> None:
        """Mark this node and all of its children as changed"""

        def visit_fn(node, depth, data):
            node.set_changed()

        self.visit_inorder(visit_fn)

    def with_color(self, text: str, style="bright") -> str:
        """Render a string that is colored if something has changed"""
        if self._rendering_change is True and self._changed is True:
            return color(text, fore=self.color, style=style)
        return text

    def add_class(self, classes) -> "MathExpression":
        """Associate a class name with an expression. This class name will be
        attached to nodes when the expression is converted to a capable output
        format.  

        See #MathExpression.to_math_ml_fragment"""
        if type(classes) == str:
            classes = [classes]
        self.classes = list(set(self.classes).union(classes))
        return self

    def to_list(self, visit: str = "preorder") -> List["MathExpression"]:
        """Convert this node hierarchy into a list."""
        results = []

        def visit_fn(node, depth, data):
            return results.append(node)

        if visit == "inorder":
            self.visit_inorder(visit_fn)
        elif visit == "preorder":
            self.visit_preorder(visit_fn)
        elif visit == "postorder":
            self.visit_postorder(visit_fn)
        else:
            raise ValueError(f"invalid visit order: {visit}")
        return results

    def clear_classes(self) -> None:
        """Clear all the classes currently set on the nodes in this expression."""

        def visit_fn(node, depth, data):
            node.classes = []

        self.visit_inorder(visit_fn)

    def find_type(self, instanceType: Type[NodeType]) -> List[NodeType]:
        """Find an expression in this tree by type.

        - instanceType: The type to check for instances of
        
        Returns the found #MathExpression objects of the given type.
        """
        results = []

        def visit_fn(node, depth, data):
            if isinstance(node, instanceType):
                return results.append(node)

        self.visit_inorder(visit_fn)
        return results

    def find_id(self, id: str) -> Optional["MathExpression"]:
        """Find an expression by its unique ID.

        Returns: The found #MathExpression or `None`
        """
        result: Optional[MathExpression] = None

        def visit_fn(node, depth, data):
            nonlocal result
            if node.id == id:
                result = node
                return STOP

        self.visit_inorder(visit_fn)
        return result

    def to_math_ml_fragment(self) -> str:
        """Convert this single node into MathML."""
        return ""

    def to_math_ml(self) -> str:
        """Convert this expression into a MathML container."""
        return "\n".join(
            [
                "<math xmlns='http:#www.w3.org/1998/Math/MathML'>",
                self.to_math_ml_fragment(),
                "</math>",
            ]
        )

    def make_ml_tag(self, tag: str, content: str, classes: List[str] = []) -> str:
        """Make a MathML tag for the given content while respecting the node's given
        classes.

        # Arguments
        tag (str): The ML tag name to create.
        content (str): The ML content to place inside of the tag.
        classes (List[str]) An array of classes to attach to this tag.

        # Returns
        (str): A MathML element with the given tag, content, and classes
        """
        classes_attr = ""
        if len(classes) > 0:
            classes_attr = f' class="{" ".join(classes)}"'
        return f"<{tag}{classes_attr}>{content}</{tag}>"

    def path_to_root(self) -> str:
        """Generate a namespaced path key to from the current node to the root.
        This key can be used to identify a node inside of a tree."""
        points = []

        def path_mark(node):
            points.append(node.__class__.__name__)

        node = self
        path_mark(node)
        while node.parent:
            node = node.parent
            path_mark(node)
        return ".".join(points)

    def clone_from_root(self, node: "MathExpression" = None) -> "MathExpression":
        """Clone this node including the entire parent hierarchy that it has. This
        is useful when you want to clone a subtree and still maintain the overall
        hierarchy.

        # Arguments
        node (MathExpression): The node to clone.

        # Returns
        (MathExpression): The cloned node.
        """
        node = node if node is not None else self
        self.cloned_node = None
        self.cloned_target = node.path_to_root()
        result = node.get_root().clone()
        if not self.cloned_node:  # pragma: nocover
            print("While cloning root of: {}".format(node))
            print(" Which is this       : {}".format(node.get_root()))
            print("Did not set the clone: {}".format(self.cloned_node))
            raise Exception("cloning root hierarchy did not clone this node")

        result = self.cloned_node
        self.cloned_node = None
        self.cloned_target = None
        return result

    def clone(self) -> "MathExpression":
        """A specialization of the clone method that can track and report a cloned
        subtree node. 
        
        See #MathExpression.clone_from_root for more details."""
        result = super().clone()
        if self.cloned_target is not None and self.path_to_root() == self.cloned_target:
            self.cloned_node = result

        return result


class UnaryExpression(MathExpression):
    """An expression that operates on one sub-expression"""

    def __init__(self, child: MathExpression = None, child_on_left: bool = True):
        super().__init__()
        self.child = child
        self.left_child = child_on_left
        self.set_child(child)

    def set_child(self, child: MathExpression = None) -> MathExpression:
        if self.left_child:
            return self.set_left(child)  # type:ignore
        else:
            return self.set_right(child)  # type:ignore

    def get_child(self) -> Optional[MathExpression]:
        if self.left_child:
            return self.left
        else:
            return self.right

    def evaluate(self, context: Dict[str, Union[float, int]] = None) -> float:
        child = self.get_child()
        if child is None:
            raise ValueError("cannot evaluate unary expression without a valid child")
        return self.operate(child.evaluate(context))

    def operate(self, value: Union[float, int]) -> Union[float, int]:
        raise NotImplementedError("Must be implemented in subclass")


# ### Negation


class NegateExpression(UnaryExpression):
    """Negate an expression, e.g. `4` becomes `-4`"""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["negate"]

    @property
    def name(self) -> str:
        return "-"

    def operate(self, value: Union[float, int]) -> Union[float, int]:
        return -value

    def __str__(self) -> str:
        return self.with_color("-{}".format(self.get_child()))

    def to_math_ml_fragment(self) -> str:
        """Convert this single node into MathML."""
        return f"-{super().to_math_ml_fragment()}"


# ### Function


class FunctionExpression(UnaryExpression):
    """A Specialized UnaryExpression that is used for functions.  The function name in
    text (used by the parser and tokenizer) is derived from the name() method on the
    class."""

    @property
    def name(self) -> str:
        raise NotImplementedError(
            "Must be implemented in subclass. Function is an abstract node"
        )

    def __str__(self) -> str:
        child = self.get_child()
        output = self.name
        if child:
            output = "{}({})".format(self.name, child)
        return self.with_color(output)


# ## Binary Expressions


class BinaryExpression(MathExpression):
    """An expression that operates on two sub-expressions"""

    def __init__(self, left=None, right=None):
        super().__init__(left=left, right=right)

    def evaluate(
        self, context: Dict[str, Union[float, int]] = None
    ) -> Union[float, int]:
        left, right = self._check()
        return self.operate(left.evaluate(context), right.evaluate(context))

    @property
    def name(self) -> str:
        raise NotImplementedError("Must be implemented in subclass")

    def get_ml_name(self) -> str:
        return self.name

    def operate(
        self, one: Union[float, int], two: Union[float, int]
    ) -> Union[float, int]:
        raise NotImplementedError("Must be implemented in subclass")

    def _check(self) -> Tuple[MathExpression, MathExpression]:
        if self.left is None or self.right is None:
            raise ValueError(
                "{}: left/right children must both be valid".format(
                    self.__class__.__name__
                )
            )
        return self.left, self.right

    def get_priority(self) -> int:
        """Return a number representing the order of operations priority
        of this node.  This can be used to check if a node is `locked`
        with respect to another node, i.e. the other node must be resolved
        first during evaluation because of it's priority.
        """
        if isinstance(self, EqualExpression):
            priority = OOO_INVALID

        if isinstance(self, AddExpression) or isinstance(self, SubtractExpression):
            priority = OOO_ADDSUB

        if isinstance(self, MultiplyExpression) or isinstance(self, DivideExpression):
            priority = OOO_MULTDIV

        if isinstance(self, PowerExpression):
            priority = OOO_EXPONENT

        if isinstance(self, FunctionExpression):
            priority = OOO_FUNCTION

        return priority

    def self_parens(self) -> bool:
        self._check()
        """Return a boolean indicating whether this node should render itself with
        a set of enclosing parnetheses or not. This is used when serializing an
        expression, to ensure the tree maintains the proper order of operations. """
        binary_parent = cast(BinaryExpression, self.parent)
        parent_is_binary = isinstance(binary_parent, BinaryExpression)
        self_pri = self.get_priority()
        if not parent_is_binary or not binary_parent:
            return False

        parent_pri = binary_parent.get_priority()
        if parent_pri > self_pri:
            return True

        parent_side = binary_parent.get_side(self)
        # If we're a left child of a parent
        if parent_pri == self_pri:
            self_addsub = isinstance(self, (AddExpression, SubtractExpression))
            parent_addsub = isinstance(self.parent, (AddExpression, SubtractExpression))
            self_muldiv = isinstance(self, (MultiplyExpression, DivideExpression))
            parent_muldiv = isinstance(
                self.parent, (MultiplyExpression, DivideExpression)
            )
            if parent_side == "right" and self_addsub and parent_addsub:
                return True
            if parent_side == "left" and self_muldiv and parent_muldiv:
                return True
        return False

    def __str__(self) -> str:
        self._check()
        out = f"{self.left} {self.with_color(self.name)} {self.right}"
        return f"({out})" if self.self_parens() else out

    def to_math_ml_fragment(self) -> str:
        """Render this node as a MathML element fragment"""
        right, left = self._check()
        right_ml = right.to_math_ml_fragment()
        left_ml = left.to_math_ml_fragment()
        op_ml = self.make_ml_tag("mo", self.get_ml_name())
        if self.self_parens():
            return self.make_ml_tag(
                "mrow",
                "<mo>(</mo>{}{}{}<mo>)</mo>".format(left_ml, op_ml, right_ml),
                self.classes,
            )
        return self.make_ml_tag(
            "mrow", "{}{}{}".format(left_ml, op_ml, right_ml), self.classes
        )


class EqualExpression(BinaryExpression):
    """Evaluate equality of two expressions"""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["equal"]

    @property
    def name(self) -> str:
        return "="

    def operate(
        self, one: Union[float, int], two: Union[float, int]
    ) -> Union[float, int]:
        """This is where assignment of context variables might make sense. But context
        is not present in the expression's `operate` method.

        !!! warning

            TODO: Investigate this thoroughly.
        """
        raise ValueError("Unsupported operation. Euqality has no operation to perform.")


class AddExpression(BinaryExpression):
    """Add one and two"""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["add"]

    @property
    def name(self) -> str:
        return "+"

    def operate(
        self, one: Union[float, int], two: Union[float, int]
    ) -> Union[float, int]:
        return one + two


class SubtractExpression(BinaryExpression):
    """Subtract one from two"""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["subtract"]

    @property
    def name(self) -> str:
        return "-"

    def operate(
        self, one: Union[float, int], two: Union[float, int]
    ) -> Union[float, int]:
        return one - two


class MultiplyExpression(BinaryExpression):
    """Multiply one and two"""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["multiply"]

    @property
    def name(self) -> str:
        return "*"

    def get_ml_name(self) -> str:
        return "&#183;"

    def operate(
        self, one: Union[float, int], two: Union[float, int]
    ) -> Union[float, int]:
        return one * two

    def __str__(self) -> str:
        """Multiplication special cases constant*variable to output `4x` instead of
        `4 * x`"""
        if isinstance(self.left, ConstantExpression):
            # const * var
            one = isinstance(self.right, VariableExpression)
            # const * var^power
            two = isinstance(self.right, PowerExpression) and isinstance(
                self.right.left, VariableExpression
            )
            if one or two:
                return self.with_color("{}{}".format(self.left, self.right))

        return super().__str__()

    def to_math_ml_fragment(self) -> str:
        left, right = self._check()
        right_ml = right.to_math_ml_fragment()
        left_ml = left.to_math_ml_fragment()
        if isinstance(self.left, ConstantExpression):
            if isinstance(self.right, VariableExpression) or isinstance(
                self.right, PowerExpression
            ):
                return "{}{}".format(left_ml, right_ml)

        return super().to_math_ml_fragment()


class DivideExpression(BinaryExpression):
    """Divide one by two"""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["divide"]

    @property
    def name(self) -> str:
        return "/"

    def get_ml_name(self) -> str:
        return "&#247;"

    def to_math_ml_fragment(self) -> str:
        left, right = self._check()
        left_ml = left.to_math_ml_fragment()
        right_ml = right.to_math_ml_fragment()
        return f"<mfrac><mi>{left_ml}</mi><mi>{right_ml}</mi></mfrac>"

    def operate(
        self, one: Union[float, int], two: Union[float, int]
    ) -> Union[float, int]:
        if two == 0:
            return float("nan")
        else:
            return one / two


class PowerExpression(BinaryExpression):
    """Raise one to the power of two"""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["power"]

    @property
    def name(self) -> str:
        return "^"

    def to_math_ml_fragment(self) -> str:
        left, right = self._check()
        right_ml = right.to_math_ml_fragment()
        left_ml = left.to_math_ml_fragment()
        # if left is mult, enclose only right in msup
        if isinstance(self.left, MultiplyExpression):
            left_ml = self.make_ml_tag("mrow", left_ml, self.classes)

        return self.make_ml_tag("msup", "{}{}".format(left_ml, right_ml), self.classes)

    def operate(
        self, one: Union[float, int], two: Union[float, int]
    ) -> Union[float, int]:
        return np.power(one, two)

    def __str__(self):
        return "{}{}{}".format(self.left, self.with_color(self.name), self.right)


class ConstantExpression(MathExpression):
    """A Constant value node, where the value is accessible as `node.value`"""

    value: Union[float, int]

    @property
    def name(self):
        if self.value % 1 == 0:
            return f"{int(self.value)}"
        return np.format_float_positional(self.value, trim="-")

    @property
    def type_id(self):
        return MathTypeKeys["constant"]

    def __init__(self, value=None):
        super().__init__()
        self.value = value

    def clone(self):
        result = super().clone()
        result.value = self.value
        return result

    def evaluate(self, context: Dict[str, Union[float, int]] = None):
        return self.value

    def __str__(self) -> str:
        return self.with_color(self.name)

    def to_math_ml_fragment(self) -> str:
        return self.make_ml_tag("mn", str(self.value), self.classes)


class VariableExpression(MathExpression):
    identifier: Optional[str]

    @property
    def name(self) -> str:
        return f"{self.identifier}"

    @property
    def type_id(self) -> int:
        id = f"_{self.identifier.lower()[0]}" if self.identifier is not None else ""
        return MathTypeKeys[f"variable{id}"]

    def __init__(self, identifier: str = None):
        super().__init__()
        self.identifier = identifier

    def clone(self) -> "VariableExpression":
        result = cast(VariableExpression, super().clone())
        result.identifier = self.identifier
        return result

    def _check(self):
        if self.identifier is None:
            raise ValueError("identifier must be a letter")

    def __str__(self) -> str:
        self._check()
        return self.with_color("{}".format(self.identifier))

    def to_math_ml_fragment(self) -> str:
        self._check()
        return self.make_ml_tag("mi", str(self.identifier), self.classes)

    def evaluate(self, context: Dict[str, Union[int, float]] = None):
        self._check()
        id = cast(str, self.identifier)
        if context and context.get(id, None) is not None:
            return context[id]

        raise ValueError(
            "cannot evaluate statement with None variable: {}".format(self.identifier)
        )


class AbsExpression(FunctionExpression):
    """Evaluates the absolute value of an expression."""

    @property
    def type_id(self) -> int:
        return MathTypeKeys["abs"]

    @property
    def name(self) -> str:
        return "abs"

    def operate(self, value: Union[float, int]) -> Union[float, int]:
        return np.absolute(value)


class SgnExpression(FunctionExpression):
    @property
    def type_id(self):
        return MathTypeKeys["sgn"]

    @property
    def name(self):
        return "sgn"

    def operate(self, value: Union[float, int]) -> Union[float, int]:
        """Determine the sign of an value.

        # Returns
        (int): -1 if negative, 1 if positive, 0 if 0"""
        if value < 0:
            return -1

        if value > 0:
            return 1

        return 0



import math
import os
import queue
import threading
import time
from multiprocessing import Queue
from typing import Any, Dict, List, Optional, Tuple, Union, cast

import gym
import numpy as np
import tensorflow as tf
from wasabi import msg

from ...envs.gym.mathy_gym_env import MathyGymEnv
from ...state import (
    MathyEnvState,
    MathyObservation,
    MathyWindowObservation,
    ObservationFeatureIndices,
    observations_to_window,
)
from ...teacher import Teacher
from ...util import discount, print_error
from .. import action_selectors
from ..episode_memory import EpisodeMemory
from ..mcts import MCTS
from ..policy_value_model import PolicyValueModel, get_or_create_policy_model
from ..trfl import discrete_policy_entropy_loss, td_lambda
from .config import A3CConfig
from .util import record, truncate, EpisodeLosses


class A3CWorker(threading.Thread):

    args: A3CConfig

    # <GLOBAL_VARS>
    global_episode = 0
    global_moving_average_reward = 0
    save_every_n_episodes = 50
    request_quit = False
    save_lock = threading.Lock()
    # </GLOBAL_VARS>

    envs: Dict[str, Any]

    losses: EpisodeLosses

    def __init__(
        self,
        args: A3CConfig,
        action_size: int,
        global_model: PolicyValueModel,
        optimizer,
        greedy_epsilon: Union[float, List[float]],
        result_queue: Queue,
        worker_idx: int,
        writer: tf.summary.SummaryWriter,
        teacher: Teacher,
        env_extra: dict,
    ):
        super(A3CWorker, self).__init__()
        self.args = args
        self.env_extra = env_extra
        self.greedy_epsilon = greedy_epsilon
        self.iteration = 0
        self.action_size = action_size
        self.result_queue = result_queue
        self.global_model = global_model
        self.optimizer = optimizer
        self.worker_idx = worker_idx
        self.teacher = teacher
        self.losses = EpisodeLosses()

        with msg.loading(f"Worker {worker_idx} starting..."):
            first_env = self.teacher.get_env(self.worker_idx, self.iteration)
            self.writer = writer
            self.local_model = get_or_create_policy_model(
                args, self.action_size, env=gym.make(first_env, **self.env_extra).mathy
            )
            self.last_model_write = -1
            self.last_histogram_write = -1
        display_e = self.greedy_epsilon
        if self.worker_idx == 0 and self.args.main_worker_use_epsilon is False:
            display_e = 0.0
        msg.good(f"Worker {worker_idx} started. (e={display_e:.3f})")

    @property
    def tb_prefix(self) -> str:
        if self.worker_idx == 0:
            return "agent"
        return f"workers/{self.worker_idx}"

    @property
    def epsilon(self) -> float:
        """Return an exploration epsilon for use in an episode"""
        e = 0.0
        if self.worker_idx == 0 and self.args.main_worker_use_epsilon is False:
            return e

        if isinstance(self.greedy_epsilon, list):
            e = np.random.choice(self.greedy_epsilon)
        elif isinstance(self.greedy_epsilon, float):
            e = self.greedy_epsilon
        else:
            raise ValueError("greedy_epsilon must either be a float or list of floats")
        e = truncate(e)
        return e

    def run(self):
        if self.args.profile:
            import cProfile

            pr = cProfile.Profile()
            pr.enable()

        episode_memory = EpisodeMemory()
        while (
            A3CWorker.global_episode < self.args.max_eps
            and A3CWorker.request_quit is False
        ):
            reward = self.run_episode(episode_memory)
            if (
                A3CWorker.global_episode
                > self.args.teacher_start_evaluations_at_episode
            ):
                win_pct = self.teacher.report_result(self.worker_idx, reward)
                if win_pct is not None:
                    with self.writer.as_default():
                        student = self.teacher.students[self.worker_idx]
                        step = self.global_model.optimizer.iterations
                        if self.worker_idx == 0:
                            tf.summary.scalar(
                                f"win_rate/{student.topic}", data=win_pct, step=step
                            )

            self.iteration += 1
            # TODO: Make this a subprocess? Python threads won't scale up well to
            #       many cores, I think.

        if self.args.profile:
            profile_name = f"worker_{self.worker_idx}.profile"
            profile_path = os.path.join(self.args.model_dir, profile_name)
            pr.disable()
            pr.dump_stats(profile_path)
            if self.args.verbose:
                print(f"PROFILER: saved {profile_path}")
        self.result_queue.put(None)

    def build_episode_selector(
        self, env: MathyGymEnv
    ) -> "action_selectors.ActionSelector":
        if self.worker_idx == 0:
            # disable dirichlet noise in worker_0
            epsilon = 0.0
        else:
            # explore based on eGreedy param (wild guess for values)
            epsilon = 0.1 + self.epsilon
        selector: action_selectors.ActionSelector
        if self.args.action_strategy == "mcts_worker_0":
            mcts = MCTS(
                env=env.mathy,
                model=self.local_model,
                num_mcts_sims=self.args.mcts_sims,
                epsilon=epsilon,
            )
            if self.worker_idx == 0:
                selector = action_selectors.MCTSActionSelector(
                    model=self.global_model,
                    worker_id=self.worker_idx,
                    mcts=mcts,
                    episode=A3CWorker.global_episode,
                )
            else:
                selector = action_selectors.A3CEpsilonGreedyActionSelector(
                    model=self.global_model,
                    worker_id=self.worker_idx,
                    epsilon=self.epsilon,
                    episode=A3CWorker.global_episode,
                )
        elif self.args.action_strategy == "mcts_worker_n":
            mcts = MCTS(
                env=env.mathy,
                model=self.local_model,
                num_mcts_sims=self.args.mcts_sims,
                epsilon=epsilon,
            )
            if self.worker_idx != 0:
                selector = action_selectors.MCTSActionSelector(
                    model=self.global_model,
                    worker_id=self.worker_idx,
                    mcts=mcts,
                    episode=A3CWorker.global_episode,
                )
            else:
                selector = action_selectors.A3CEpsilonGreedyActionSelector(
                    model=self.global_model,
                    worker_id=self.worker_idx,
                    epsilon=self.epsilon,
                    episode=A3CWorker.global_episode,
                )
        elif self.args.action_strategy in ["a3c"]:
            selector = action_selectors.A3CEpsilonGreedyActionSelector(
                model=self.global_model,
                worker_id=self.worker_idx,
                epsilon=self.epsilon,
                episode=A3CWorker.global_episode,
            )
        else:
            raise EnvironmentError(
                f"Unknown action_strategy: {self.args.action_strategy}"
            )
        return selector

    def run_episode(self, episode_memory: EpisodeMemory) -> float:
        env_name = self.teacher.get_env(self.worker_idx, self.iteration)
        env = gym.make(env_name, **self.env_extra)
        episode_memory.clear()
        self.ep_loss = 0
        ep_reward = 0.0
        ep_steps = 1
        time_count = 0
        done = False
        last_observation: MathyObservation = env.reset()
        last_text = env.state.agent.problem
        last_action = -1
        last_reward = 0.0

        selector = self.build_episode_selector(env)

        while not done and A3CWorker.request_quit is False:
            if self.args.print_training and self.worker_idx == 0:
                env.render(last_action=last_action, last_reward=last_reward)
            window = episode_memory.to_window_observation(
                last_observation, window_size=self.args.prediction_window_size
            )
            try:
                action, value = selector.select(
                    last_state=env.state,
                    last_window=window,
                    last_action=last_action,
                    last_reward=last_reward,
                )
            except BaseException as err:
                print_error(err, "failed to select an action during an episode step")
                continue

            # Take an env step
            observation, reward, done, last_obs_info = env.step(action)
            ep_reward += reward
            episode_memory.store(
                observation=last_observation, action=action, reward=reward, value=value,
            )
            if time_count == self.args.update_gradients_every or done:
                if done and self.args.print_training and self.worker_idx == 0:
                    env.render(last_action=last_action, last_reward=last_reward)

                self.update_global_network(done, observation, episode_memory)
                self.maybe_write_histograms()
                time_count = 0
                if done:
                    self.finish_episode(
                        last_obs_info.get("win", False), ep_reward, ep_steps, env.state
                    )

            ep_steps += 1
            time_count += 1
            last_observation = observation
            last_action = int(action)
            last_reward = reward

            # If there are multiple workers, apply a worker sleep
            # to give the system some breathing room.
            if self.args.num_workers > 1:
                # The greedy worker sleeps for a shorter period of time
                sleep = self.args.worker_wait
                if self.worker_idx == 0:
                    sleep = max(sleep // 100, 0.005)
                # Workers wait between each step so that it's possible
                # to run more workers than there are CPUs available.
                time.sleep(sleep)
        return ep_reward

    def maybe_write_episode_summaries(
        self, episode_reward: float, episode_steps: int, last_state: MathyEnvState
    ):
        assert self.worker_idx == 0, "only write summaries for greedy worker"
        # Track metrics for all workers
        name = self.teacher.get_env(self.worker_idx, self.iteration)
        step = self.global_model.optimizer.iterations
        with self.writer.as_default():
            agent_state = last_state.agent
            steps = int(last_state.max_moves - agent_state.moves_remaining)
            rwd = truncate(episode_reward)
            p_text = f"{agent_state.history[0].raw} = {agent_state.history[-1].raw}"
            outcome = "SOLVED" if episode_reward > 0.0 else "FAILED"
            out_text = f"{outcome} [steps: {steps}, reward: {rwd}]: {p_text}"
            tf.summary.text(f"{name}/summary", data=out_text, step=step)

            # Track global model metrics
            tf.summary.scalar(
                f"agent/mean_episode_reward",
                data=A3CWorker.global_moving_average_reward,
                step=step,
            )

    def maybe_write_histograms(self) -> None:
        if self.worker_idx != 0:
            return
        step = self.global_model.optimizer.iterations.numpy()
        next_write = self.last_histogram_write + self.args.summary_interval
        if step >= next_write or self.last_histogram_write == -1:
            with self.writer.as_default():
                self.last_histogram_write = step
                for var in self.local_model.trainable_variables:
                    tf.summary.histogram(
                        var.name, var, step=self.global_model.optimizer.iterations,
                    )

    def update_global_network(
        self, done: bool, observation: MathyObservation, episode_memory: EpisodeMemory,
    ):
        # Calculate gradient wrt to local model. We do so by tracking the
        # variables involved in computing the loss by using tf.GradientTape
        with tf.GradientTape() as tape:
            loss_tuple = self.compute_loss(
                done=done,
                observation=observation,
                episode_memory=episode_memory,
                gamma=self.args.gamma,
            )
            (
                pi_loss,
                value_loss,
                rp_loss,
                entropy_loss,
                aux_losses,
                total_loss,
            ) = loss_tuple

        self.losses.increment("loss", total_loss)
        self.losses.increment("pi", pi_loss)
        self.losses.increment("r", rp_loss)
        self.losses.increment("v", value_loss)
        self.losses.increment("h", entropy_loss)
        for k in aux_losses.keys():
            self.losses.increment(k, aux_losses[k].numpy())

        # Calculate local gradients
        grads = tape.gradient(total_loss, self.local_model.trainable_weights)
        # Push local gradients to global model

        zipped_gradients = zip(grads, self.global_model.trainable_weights)
        # Assert that we always have some gradient flow in each trainable var

        # TODO: Make this a unit test. It degrades performance at train time
        # for grad, var in zipped_gradients:
        #     nonzero_grads = tf.math.count_nonzero(grad).numpy()
        #     grad_sum = tf.math.reduce_sum(grad).numpy()
        #     # if "lstm" in var.name and self.worker_idx == 0:
        #     #     print(f"[{var.name}] {grad_sum}")
        #     if nonzero_grads == 0:
        #         tf.print(grad_sum)
        #         raise ValueError(f"{var.name} has no gradient")

        self.optimizer.apply_gradients(zipped_gradients)
        # Update local model with new weights
        self.local_model.set_weights(self.global_model.get_weights())

        if done:
            episode_memory.clear()
        else:
            episode_memory.clear_except_window(self.args.prediction_window_size)

    def finish_episode(
        self,
        is_win: bool,
        episode_reward: float,
        episode_steps: int,
        last_state: MathyEnvState,
    ):
        env_name = self.teacher.get_env(self.worker_idx, self.iteration)

        # Only observe/track the most-greedy worker (high epsilon exploration
        # stats are unlikely to be consistent or informative)
        if self.worker_idx == 0:
            A3CWorker.global_moving_average_reward = record(
                A3CWorker.global_episode,
                is_win,
                episode_reward,
                self.worker_idx,
                A3CWorker.global_moving_average_reward,
                self.result_queue,
                self.losses,
                episode_steps,
                env_name,
            )
            self.maybe_write_episode_summaries(
                episode_reward, episode_steps, last_state
            )

            step = self.global_model.optimizer.iterations.numpy()
            next_write = self.last_model_write + A3CWorker.save_every_n_episodes
            if step >= next_write or self.last_model_write == -1:
                self.last_model_write = step
                self.write_global_model()

        A3CWorker.global_episode += 1
        self.losses.reset()

    def write_global_model(self, increment_episode=True):
        with A3CWorker.save_lock:
            # Do this inside the lock so other threads can't also acquire the
            # lock in the time between when it's released and assigned outside
            # of the if conditional.
            if increment_episode is True:
                A3CWorker.global_episode += 1
                self.global_model.save()

    def compute_policy_value_loss(
        self,
        done: bool,
        observation: MathyObservation,
        episode_memory: EpisodeMemory,
        gamma=0.99,
    ):
        step = self.global_model.optimizer.iterations
        if done:
            bootstrap_value = 0.0  # terminal
        else:
            # Predict the reward using the local network
            _, values, _, _, _ = self.local_model.call(
                observations_to_window([observation]).to_inputs()
            )
            # Select the last timestep
            values = values[-1]
            bootstrap_value = tf.squeeze(values).numpy()

        discounted_rewards: List[float] = []
        for reward in episode_memory.rewards[::-1]:
            bootstrap_value = reward + gamma * bootstrap_value
            discounted_rewards.append(bootstrap_value)
        discounted_rewards.reverse()
        discounted_rewards = tf.convert_to_tensor(
            value=np.array(discounted_rewards)[:, None], dtype=tf.float32
        )

        batch_size = len(episode_memory.actions)
        sequence_length = len(episode_memory.observations[0].nodes)
        inputs = episode_memory.to_episode_window().to_inputs()
        model_results = self.local_model.call(inputs)
        logits, values, trimmed_logits, reward_logits, attentions = model_results

        logits = tf.reshape(logits, [batch_size, -1])

        # Calculate entropy and policy loss
        h_loss = discrete_policy_entropy_loss(
            logits, normalise=self.args.normalize_entropy_loss
        )
        # Scale entropy loss down
        entropy_loss = h_loss.loss * self.args.entropy_loss_scaling
        entropy_loss = tf.reduce_mean(entropy_loss)

        rewards_tensor = tf.convert_to_tensor(episode_memory.rewards, dtype=tf.float32)
        rewards_tensor = tf.expand_dims(rewards_tensor, 1)
        rp_loss = tf.reduce_mean(tf.keras.losses.MSE(rewards_tensor, reward_logits))
        pcontinues = tf.convert_to_tensor([[gamma]] * batch_size, dtype=tf.float32)
        bootstrap_value = tf.convert_to_tensor([bootstrap_value], dtype=tf.float32)

        lambda_loss = td_lambda(
            state_values=values,
            rewards=rewards_tensor,
            pcontinues=pcontinues,
            bootstrap_value=bootstrap_value,
            lambda_=self.args.td_lambda,
        )
        advantage = lambda_loss.extra.temporal_differences
        # Value loss
        value_loss = tf.reduce_mean(lambda_loss.loss)

        # Policy Loss
        policy_loss = tf.nn.sparse_softmax_cross_entropy_with_logits(
            labels=episode_memory.actions, logits=logits
        )

        policy_loss *= advantage
        policy_loss = tf.reduce_mean(policy_loss)
        if self.args.normalize_pi_loss:
            policy_loss /= sequence_length
        # Scale the policy loss down by the seq_len to make invariant to length
        total_loss = value_loss + policy_loss + entropy_loss + rp_loss
        prefix = self.tb_prefix
        tf.summary.scalar(f"losses/{prefix}/policy_loss", data=policy_loss, step=step)
        tf.summary.scalar(f"losses/{prefix}/value_loss", data=value_loss, step=step)
        tf.summary.scalar(f"losses/{prefix}/entropy_loss", data=entropy_loss, step=step)
        tf.summary.scalar(f"losses/{prefix}/rp_loss", data=rp_loss, step=step)
        tf.summary.scalar(
            f"settings/learning_rate", data=self.optimizer.lr(step), step=step
        )
        return (
            (
                policy_loss,
                value_loss,
                rp_loss,
                entropy_loss,
                total_loss,
                discounted_rewards,
            ),
            model_results,
        )

    def compute_loss(
        self,
        *,
        done: bool,
        observation: MathyObservation,
        episode_memory: EpisodeMemory,
        gamma=0.99,
    ):
        with self.writer.as_default():
            step = self.global_model.optimizer.iterations
            loss_tuple, model_results = self.compute_policy_value_loss(
                done, observation, episode_memory
            )
            (
                pi_loss,
                v_loss,
                rp_loss,
                h_loss,
                total_loss,
                discounted_rewards,
            ) = loss_tuple
            aux_losses = {}
            aux_weight = self.args.aux_tasks_weight_scale
            for key in aux_losses.keys():
                tf.summary.scalar(
                    f"{self.tb_prefix}/{key}_loss", data=aux_losses[key], step=step
                )

            tf.summary.scalar(
                f"{self.tb_prefix}/total_loss", data=total_loss, step=step
            )

        return pi_loss, v_loss, rp_loss, h_loss, aux_losses, total_loss



# Copyright 2018 The trfl Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or  implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Ops for computing v-trace learning targets."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

# Dependency imports
import tensorflow as tf

from tensorflow.python.util import nest

VTraceFromLogitsReturns = collections.namedtuple(
    "VTraceFromLogitsReturns",
    [
        "vs",
        "pg_advantages",
        "log_rhos",
        "behaviour_action_log_probs",
        "target_action_log_probs",
    ],
)
VTraceReturns = collections.namedtuple("VTraceReturns", "vs pg_advantages")


def log_probs_from_logits_and_actions(policy_logits, actions):
    """Computes action log-probs from policy logits and actions.

  In the notation used throughout documentation and comments, T refers to the
  time dimension ranging from 0 to T-1. B refers to the batch size and
  NUM_ACTIONS refers to the number of actions.

  Args:
    policy_logits: A float32 tensor of shape `[T, B, NUM_ACTIONS]` with
      un-normalized log-probabilities parameterizing a softmax policy.
    actions: An int32 tensor of shape `[T, B]` with actions.

  Returns:
    A float32 tensor of shape `[T, B]` corresponding to the sampling log
    probability of the chosen action w.r.t. the policy.
  """
    policy_logits = tf.convert_to_tensor(value=policy_logits, dtype=tf.float32)
    actions = tf.convert_to_tensor(value=actions, dtype=tf.int32)

    policy_logits.shape.assert_has_rank(3)
    actions.shape.assert_has_rank(2)

    return -tf.nn.sparse_softmax_cross_entropy_with_logits(
        logits=policy_logits, labels=actions
    )


def vtrace_from_logits(
    behaviour_policy_logits,
    target_policy_logits,
    actions,
    discounts,
    rewards,
    values,
    bootstrap_value,
    clip_rho_threshold=1.0,
    clip_pg_rho_threshold=1.0,
    name="vtrace_from_logits",
):
    r"""V-trace for softmax policies.

  Calculates V-trace actor critic targets for softmax polices as described in

  "IMPALA: Scalable Distributed Deep-RL with
  Importance Weighted Actor-Learner Architectures"
  by Espeholt, Soyer, Munos et al.

  Target policy refers to the policy we are interested in improving and
  behaviour policy refers to the policy that generated the given
  rewards and actions.

  In the notation used throughout documentation and comments, `T` refers to the
  time dimension ranging from `0` to `T-1`. `B` refers to the batch size and
  `NUM_ACTIONS` refers to the number of actions.

  Args:
    behaviour_policy_logits: A float32 tensor of shape `[T, B, NUM_ACTIONS]`
      with un-normalized log-probabilities parametrizing the softmax behaviour
      policy.
    target_policy_logits: A float32 tensor of shape `[T, B, NUM_ACTIONS]` with
      un-normalized log-probabilities parametrizing the softmax target policy.
    actions: An int32 tensor of shape `[T, B]` of actions sampled from the
      behaviour policy.
    discounts: A float32 tensor of shape `[T, B]` with the discount encountered
      when following the behaviour policy.
    rewards: A float32 tensor of shape `[T, B]` with the rewards generated by
      following the behaviour policy.
    values: A float32 tensor of shape `[T, B]` with the value function estimates
      wrt. the target policy.
    bootstrap_value: A float32 of shape `[B]` with the value function estimate
      at time T.
    clip_rho_threshold: A scalar float32 tensor with the clipping threshold for
      importance weights (rho) when calculating the baseline targets (vs).
      rho^bar in the paper.
    clip_pg_rho_threshold: A scalar float32 tensor with the clipping threshold
      on rho_s in \rho_s \delta log \pi(a|x) (r + \gamma v_{s+1} - V(x_s)).
    name: The name scope that all V-trace operations will be created in.

  Returns:
    A `VTraceFromLogitsReturns` namedtuple with the following fields:
      vs: A float32 tensor of shape `[T, B]`. Can be used as target to train a
          baseline (V(x_t) - vs_t)^2.
      pg_advantages: A float 32 tensor of shape `[T, B]`. Can be used as an
        estimate of the advantage in the calculation of policy gradients.
      log_rhos: A float32 tensor of shape `[T, B]` containing the log importance
        sampling weights (log rhos).
      behaviour_action_log_probs: A float32 tensor of shape `[T, B]` containing
        behaviour policy action log probabilities (log \mu(a_t)).
      target_action_log_probs: A float32 tensor of shape `[T, B]` containing
        target policy action probabilities (log \pi(a_t)).
  """
    behaviour_policy_logits = tf.convert_to_tensor(
        value=behaviour_policy_logits, dtype=tf.float32
    )
    target_policy_logits = tf.convert_to_tensor(
        value=target_policy_logits, dtype=tf.float32
    )
    actions = tf.convert_to_tensor(value=actions, dtype=tf.int32)

    # Make sure tensor ranks are as expected.
    # The rest will be checked by from_action_log_probs.
    behaviour_policy_logits.shape.assert_has_rank(3)
    target_policy_logits.shape.assert_has_rank(3)
    actions.shape.assert_has_rank(2)

    with tf.compat.v1.name_scope(
        name,
        values=[
            behaviour_policy_logits,
            target_policy_logits,
            actions,
            discounts,
            rewards,
            values,
            bootstrap_value,
        ],
    ):
        target_action_log_probs = log_probs_from_logits_and_actions(
            target_policy_logits, actions
        )
        behaviour_action_log_probs = log_probs_from_logits_and_actions(
            behaviour_policy_logits, actions
        )
        log_rhos = target_action_log_probs - behaviour_action_log_probs
        vtrace_returns = vtrace_from_importance_weights(
            log_rhos=log_rhos,
            discounts=discounts,
            rewards=rewards,
            values=values,
            bootstrap_value=bootstrap_value,
            clip_rho_threshold=clip_rho_threshold,
            clip_pg_rho_threshold=clip_pg_rho_threshold,
        )
        return VTraceFromLogitsReturns(
            log_rhos=log_rhos,
            behaviour_action_log_probs=behaviour_action_log_probs,
            target_action_log_probs=target_action_log_probs,
            **vtrace_returns._asdict()
        )


def vtrace_from_importance_weights(
    log_rhos,
    discounts,
    rewards,
    values,
    bootstrap_value,
    clip_rho_threshold=1.0,
    clip_pg_rho_threshold=1.0,
    name="vtrace_from_importance_weights",
):
    r"""V-trace from log importance weights.

  Calculates V-trace actor critic targets as described in

  "IMPALA: Scalable Distributed Deep-RL with
  Importance Weighted Actor-Learner Architectures"
  by Espeholt, Soyer, Munos et al.

  In the notation used throughout documentation and comments, T refers to the
  time dimension ranging from 0 to T-1. B refers to the batch size and
  NUM_ACTIONS refers to the number of actions. This code also supports the
  case where all tensors have the same number of additional dimensions, e.g.,
  `rewards` is `[T, B, C]`, `values` is `[T, B, C]`, `bootstrap_value`
  is `[B, C]`.

  Args:
    log_rhos: A float32 tensor of shape `[T, B, NUM_ACTIONS]` representing the
      log importance sampling weights, i.e.
      log(target_policy(a) / behaviour_policy(a)). V-trace performs operations
      on rhos in log-space for numerical stability.
    discounts: A float32 tensor of shape `[T, B]` with discounts encountered
      when following the behaviour policy.
    rewards: A float32 tensor of shape `[T, B]` containing rewards generated by
      following the behaviour policy.
    values: A float32 tensor of shape `[T, B]` with the value function estimates
      wrt. the target policy.
    bootstrap_value: A float32 of shape `[B]` with the value function estimate
      at time T.
    clip_rho_threshold: A scalar float32 tensor with the clipping threshold for
      importance weights (rho) when calculating the baseline targets (vs).
      rho^bar in the paper. If None, no clipping is applied.
    clip_pg_rho_threshold: A scalar float32 tensor with the clipping threshold
      on rho_s in \rho_s \delta log \pi(a|x) (r + \gamma v_{s+1} - V(x_s)). If
      None, no clipping is applied.
    name: The name scope that all V-trace operations will be created in.

  Returns:
    A VTraceReturns namedtuple (vs, pg_advantages) where:
      vs: A float32 tensor of shape `[T, B]`. Can be used as target to
        train a baseline (V(x_t) - vs_t)^2.
      pg_advantages: A float32 tensor of shape `[T, B]`. Can be used as the
        advantage in the calculation of policy gradients.
  """
    log_rhos = tf.convert_to_tensor(value=log_rhos, dtype=tf.float32)
    discounts = tf.convert_to_tensor(value=discounts, dtype=tf.float32)
    rewards = tf.convert_to_tensor(value=rewards, dtype=tf.float32)
    values = tf.convert_to_tensor(value=values, dtype=tf.float32)
    bootstrap_value = tf.convert_to_tensor(value=bootstrap_value, dtype=tf.float32)
    if clip_rho_threshold is not None:
        clip_rho_threshold = tf.convert_to_tensor(
            value=clip_rho_threshold, dtype=tf.float32
        )
    if clip_pg_rho_threshold is not None:
        clip_pg_rho_threshold = tf.convert_to_tensor(
            value=clip_pg_rho_threshold, dtype=tf.float32
        )

    # Make sure tensor ranks are consistent.
    rho_rank = log_rhos.shape.ndims  # Usually 2.
    values.shape.assert_has_rank(rho_rank)
    bootstrap_value.shape.assert_has_rank(rho_rank - 1)
    discounts.shape.assert_has_rank(rho_rank)
    rewards.shape.assert_has_rank(rho_rank)
    if clip_rho_threshold is not None:
        clip_rho_threshold.shape.assert_has_rank(0)
    if clip_pg_rho_threshold is not None:
        clip_pg_rho_threshold.shape.assert_has_rank(0)

    with tf.compat.v1.name_scope(
        name, values=[log_rhos, discounts, rewards, values, bootstrap_value]
    ):
        rhos = tf.exp(log_rhos)
        if clip_rho_threshold is not None:
            clipped_rhos = tf.minimum(clip_rho_threshold, rhos, name="clipped_rhos")
        else:
            clipped_rhos = rhos

        cs = tf.minimum(1.0, rhos, name="cs")
        # Append bootstrapped value to get [v1, ..., v_t+1]
        values_t_plus_1 = tf.concat(
            [values[1:], tf.expand_dims(bootstrap_value, 0)], axis=0
        )
        deltas = clipped_rhos * (rewards + discounts * values_t_plus_1 - values)

        # Note that all sequences are reversed, computation starts from the back.
        sequences = (
            tf.reverse(discounts, axis=[0]),
            tf.reverse(cs, axis=[0]),
            tf.reverse(deltas, axis=[0]),
        )
        # V-trace vs are calculated through a scan from the back to the beginning
        # of the given trajectory.
        def scanfunc(acc, sequence_item):
            discount_t, c_t, delta_t = sequence_item
            return delta_t + discount_t * c_t * acc

        initial_values = tf.zeros_like(bootstrap_value)
        vs_minus_v_xs = tf.scan(
            fn=scanfunc,
            elems=sequences,
            initializer=initial_values,
            parallel_iterations=1,
            back_prop=False,
            name="scan",
        )
        # Reverse the results back to original order.
        vs_minus_v_xs = tf.reverse(vs_minus_v_xs, [0], name="vs_minus_v_xs")

        # Add V(x_s) to get v_s.
        vs = tf.add(vs_minus_v_xs, values, name="vs")

        # Advantage for policy gradient.
        vs_t_plus_1 = tf.concat([vs[1:], tf.expand_dims(bootstrap_value, 0)], axis=0)
        if clip_pg_rho_threshold is not None:
            clipped_pg_rhos = tf.minimum(
                clip_pg_rho_threshold, rhos, name="clipped_pg_rhos"
            )
        else:
            clipped_pg_rhos = rhos
        pg_advantages = clipped_pg_rhos * (rewards + discounts * vs_t_plus_1 - values)

        # Make sure no gradients backpropagated through the returned values.
        return VTraceReturns(
            vs=tf.stop_gradient(vs), pg_advantages=tf.stop_gradient(pg_advantages)
        )




from typing import List, Tuple

import numpy as np

from ..state import MathyEnvState, MathyWindowObservation
from .policy_value_model import PolicyValueModel
from .mcts import MCTS


class ActionSelector:
    """An episode-specific selector of actions"""

    model: PolicyValueModel
    worker_id: int
    episode: int

    def __init__(self, *, model: PolicyValueModel, episode: int, worker_id: int):
        self.model = model
        self.worker_id = worker_id
        self.episode = episode

    def select(
        self,
        *,
        last_state: MathyEnvState,
        last_window: MathyWindowObservation,
        last_action: int,
        last_reward: float,
    ) -> Tuple[int, float]:
        raise NotImplementedError(self.select)


class GreedyActionSelector(ActionSelector):
    def select(
        self,
        *,
        last_state: MathyEnvState,
        last_window: MathyWindowObservation,
        last_action: int,
        last_reward: float,
    ) -> Tuple[int, float]:
        probs, value = self.model.predict_next(last_window.to_inputs())
        action = np.argmax(probs)
        return action, float(value)


class A3CEpsilonGreedyActionSelector(ActionSelector):
    def __init__(self, *, epsilon: float, **kwargs):
        super(A3CEpsilonGreedyActionSelector, self).__init__(**kwargs)
        self.epsilon = epsilon

    def select(
        self,
        *,
        last_state: MathyEnvState,
        last_window: MathyWindowObservation,
        last_action: int,
        last_reward: float,
    ) -> Tuple[int, float]:

        probs, value = self.model.predict_next(last_window.to_inputs())
        last_move_mask = last_window.mask[-1]
        no_random = bool(self.worker_id == 0)
        if not no_random and np.random.random() < self.epsilon:
            # Select a random action
            action_mask = last_move_mask[:]
            # normalize all valid actions to equal probability
            actions = action_mask / np.sum(action_mask)
            action = np.random.choice(len(actions), p=actions)
        else:
            action = np.argmax(probs)
        return action, float(value)


class MCTSActionSelector(ActionSelector):
    def __init__(self, mcts: MCTS, **kwargs):
        super(MCTSActionSelector, self).__init__(**kwargs)
        self.mcts = mcts

    def select(
        self,
        *,
        last_state: MathyEnvState,
        last_window: MathyWindowObservation,
        last_action: int,
        last_reward: float,
    ) -> Tuple[int, float]:
        probs, value = self.mcts.estimate_policy(last_state)
        return np.argmax(probs), value



from ...agents.base_config import BaseConfig


class SelfPlayConfig(BaseConfig):
    batch_size = 128
    # PyTorch models need padded batches of the max length a sequence might be.
    # Mathy expressions can occasionally go over 128 nodes, but I've never made
    # an expression more than 256. If we go over this limit, it might be time to
    # introduce the Reformer.
    max_sequence_length: int = 256
    epochs = 10
    mcts_sims: int = 50
    temperature_threshold: float = 0.5
    self_play_problems: int = 64
    training_iterations: int = 100
    cpuct: float = 1.0
    normalization_style: str = "batch"
    # When profile is true and workers == 1 the main thread will output worker_0.profile on exit
    profile: bool = False
    # Don't use the LSTM with Zero agents because the random sampling breaks
    # timestep correlations across the batch.
    use_lstm: bool = False




import random
from typing import Any, Dict, List, Optional

from pydantic import BaseModel


class StudentEvaluation(BaseModel):
    env: str
    total: int
    solved: int
    failed: int


class Topic(BaseModel):
    # The topic name used to interpolate the env name
    name: str
    # The current topic difficulty
    difficulty: str = "easy"
    # The number of observations for the current eval window
    count: int = 0
    # The current positives count
    positives: int = 0
    # The current negatives count
    negatives: int = 0

    def reset_counts(self):
        self.positives = 0
        self.negatives = 0
        self.count = 0


class Student(BaseModel):
    # The index of the student in the students list
    id: int
    # Current topic of study
    topic: str
    # The current difficulty settings for each env
    topics: Dict[str, Topic]


class Teacher:
    
    students: List[Student]

    def __init__(
        self,
        topic_names: List[str],
        num_students: int = 1,
        difficulty: Optional[str] = None,
        eval_window: int = 50,
        win_threshold: float = 0.95,
        lose_threshold: float = 0.34,
    ):
        self.topic_names = topic_names
        self.eval_window = eval_window
        self.win_threshold = win_threshold
        self.lose_threshold = lose_threshold
        self.difficulty = difficulty
        if self.difficulty is not None:
            print(f"difficulty will not adjust and is fixed to: {self.difficulty}")
        self.initialize_students(num_students)

    def initialize_students(self, num_students: int):
        self.num_students = num_students
        self.students = []
        for i in range(self.num_students):
            student_topics = {}
            start_topic = random.choice(self.topic_names)
            for topic in self.topic_names:
                difficulty = self.difficulty if self.difficulty is not None else "easy"
                student_topics[topic] = Topic(name=topic, difficulty=difficulty)
            self.students.append(
                Student(id=i, topic=start_topic, topics=student_topics)
            )

    def previous_difficulty(self, difficulty: str) -> str:
        """Return the previous difficulty level given an input difficulty.
        
        # Arguments
        difficulty (str): The difficulty to decrease

        # Returns
        (str): The difficulty level before the input, if any.
        """
        if difficulty == "hard":
            return "normal"
        elif difficulty == "normal":
            return "easy"
        return "easy"

    def next_difficulty(self, difficulty: str) -> str:
        """Return the previous difficulty level given an input difficulty.
        
        # Arguments
        difficulty (str): The difficulty to increase

        # Returns
        (str): The difficulty level after the input, if any.
        """
        if difficulty == "easy":
            return "normal"
        elif difficulty == "normal":
            return "hard"
        return "hard"

    def report_result(
        self, student_index: int, reward: float, data: Any = None
    ) -> Optional[float]:
        student = self.students[student_index]
        topic: Topic = student.topics[student.topic]
        if reward > 0.0:
            topic.positives += 1
        else:
            topic.negatives += 1
        topic.count += 1

        if topic.count >= self.eval_window:
            win_ratio = topic.positives / self.eval_window
            action = "kept at the same difficulty, to gather more experience"
            # If the difficulty is locked, don't adjust it.
            if self.difficulty is not None:
                pass
            elif win_ratio >= self.win_threshold:
                topic.difficulty = self.next_difficulty(topic.difficulty)
                action = "promoted"
            elif win_ratio <= self.lose_threshold:
                topic.difficulty = self.previous_difficulty(topic.difficulty)
                action = "demoted"
            if student_index == 0:
                pct = int(win_ratio * 100)
                type = topic.name
                diff = topic.difficulty
                print(
                    f"Solved {pct}% of {type} problems and was {action}. "
                    f"Next round will use {diff} difficulty problems."
                )
            topic.reset_counts()
            return win_ratio
        return None

    def get_env(self, student_index: int, iteration: int) -> str:
        """Get the current environment a student should be using.

        # Arguments
        student_index (int): The index of the student in `self.students` array.
        iteration (int): The current iteration (usually an episode).

        # Returns
        (str): The name of a mathy environment to use.
        """
        student = self.students[student_index]
        # The console printing student is special, it trains in everything
        len_topics = len(self.topic_names)
        student.topic = self.topic_names[iteration % len_topics]
        topic = student.topics[student.topic]
        return f"mathy-{topic.name}-{topic.difficulty}-v0"



import pytest
import tensorflow as tf

from ..mathy import envs
from ..mathy.agents.base_config import BaseConfig
from ..mathy.agents.policy_value_model import PolicyValueModel
from ..mathy.env import MathyEnv
from ..mathy.api import Mathy
from ..mathy.state import MathyObservation, observations_to_window


def test_mathy_requires_model_and_config_or_path():
    with pytest.raises(ValueError):
        mt = Mathy()


def test_mathy_policy_value_subclass_error():
    """Do not allowing entirely custom models. They must subclass PolicyValueModel"""
    # NOTE: if this test is breaking your model, let's talk about it on Github
    model = tf.keras.Model()
    with pytest.raises(ValueError):
        mt = Mathy(model=model)


def test_mathy_with_model_and_config():
    config = BaseConfig()
    model = PolicyValueModel(config, predictions=2)
    mt = Mathy(model=model, config=config)




from .binomial_distribute import BinomialDistribute
from .complex_simplify import ComplexSimplify
from .poly_simplify_blockers import PolySimplifyBlockers
from .poly_combine_in_place import PolyCombineInPlace
from .poly_commute_like_terms import PolyCommuteLikeTerms
from .poly_grouping import PolyGroupLikeTerms
from .poly_haystack_like_terms import PolyHaystackLikeTerms
from .poly_simplify import PolySimplify


MATHY_BUILTIN_ENVS = [
    BinomialDistribute,
    ComplexSimplify,
    PolySimplifyBlockers,
    PolyCombineInPlace,
    PolyCommuteLikeTerms,
    PolyGroupLikeTerms,
    PolyHaystackLikeTerms,
    PolySimplify,
]



import time
import curses
import uuid
from typing import Optional, Union

# ## Constants

# Return this from a node visit function to abort a tree visit.
STOP = "stop"
# The constant representing the left child side of a node.
LEFT = "left"
# The constant representing the right child side of a node.
RIGHT = "right"


class BinaryTreeNode:
    """
    The binary tree node is the base node for all of our trees, and provides a
    rich set of methods for constructing, inspecting, and modifying them.
    The node itself defines the structure of the binary tree, having left and right
    children, and a parent.
    """

    _idCounter = 0

    # Tree layout mutations. Thanks, 2009 Justin. :( :(
    x: Optional[float]
    y: Optional[float]
    offset: Optional[float]

    left: Optional["BinaryTreeNode"]
    right: Optional["BinaryTreeNode"]
    parent: Optional["BinaryTreeNode"]

    #  Allow specifying children in the constructor
    def __init__(
        self,
        left: "BinaryTreeNode" = None,
        right: "BinaryTreeNode" = None,
        parent: "BinaryTreeNode" = None,
        id: Optional[str] = None,
    ):
        if id is None:
            BinaryTreeNode._idCounter = BinaryTreeNode._idCounter + 1
            id = f"mn-{BinaryTreeNode._idCounter}"
        self.id = id
        self.left = None
        self.right = None
        self.set_left(left)
        self.set_right(right)
        self.parent = parent

    def clone(self):
        """Create a clone of this tree"""
        result = self.__class__()
        result.id = self.id
        if self.left:
            result.set_left(self.left.clone())

        if self.right:
            result.set_right(self.right.clone())
        return result

    def is_leaf(self):
        """Is this node a leaf?  A node is a leaf if it has no children."""
        return not self.left and not self.right

    def __str__(self):
        """Serialize the node as a str"""
        return "{} {}".format(self.left, self.right)

    @property
    def name(self):
        """Human readable name for this node."""
        return "BinaryTreeNode"

    def rotate(self):
        """
        Rotate a node, changing the structure of the tree, without modifying
        the order of the nodes in the tree.
        """
        node = self
        parent = self.parent
        if not node or not parent:
            return self

        grand_parent = parent.parent
        if node == parent.left:
            parent.set_left(node.right)
            node.right = parent
            parent.parent = node
        else:
            parent.set_right(node.left)
            node.left = parent
            parent.parent = node

        node.parent = grand_parent
        if not grand_parent:
            return self

        if parent == grand_parent.left:
            grand_parent.left = node
        else:
            grand_parent.right = node
        return self

    def visit_preorder(self, visit_fn, depth=0, data=None):
        """Visit the tree preorder, which visits the current node, then its left
        child, and then its right child.

        *Visit -> Left -> Right*

        This method accepts a function that will be invoked for each node in the
        tree.  The callback function is passed three arguments: the node being
        visited, the current depth in the tree, and a user specified data parameter.

        !!! info

            Traversals may be canceled by returning `STOP` from any visit function.
        """
        if visit_fn and visit_fn(self, depth, data) == STOP:
            return STOP

        if self.left and self.left.visit_preorder(visit_fn, depth + 1, data) == STOP:
            return STOP

        if self.right and self.right.visit_preorder(visit_fn, depth + 1, data) == STOP:
            return STOP

    def visit_inorder(self, visit_fn, depth=0, data=None):
        """Visit the tree inorder, which visits the left child, then the current node,
        and then its right child.

        *Left -> Visit -> Right*

        This method accepts a function that will be invoked for each node in the
        tree.  The callback function is passed three arguments: the node being
        visited, the current depth in the tree, and a user specified data parameter.

        !!! info

            Traversals may be canceled by returning `STOP` from any visit function.
        """
        if self.left and self.left.visit_inorder(visit_fn, depth + 1, data) == STOP:
            return STOP

        if visit_fn and visit_fn(self, depth, data) == STOP:
            return STOP

        if self.right and self.right.visit_inorder(visit_fn, depth + 1, data) == STOP:
            return STOP

    def visit_postorder(self, visit_fn, depth=0, data=None):
        """Visit the tree postorder, which visits its left child, then its right child,
        and finally the current node.

        *Left -> Right -> Visit*

        This method accepts a function that will be invoked for each node in the
        tree.  The callback function is passed three arguments: the node being
        visited, the current depth in the tree, and a user specified data parameter.

        !!! info

            Traversals may be canceled by returning `STOP` from any visit function.
        """
        if self.left and self.left.visit_postorder(visit_fn, depth + 1, data) == STOP:
            return STOP

        if self.right and self.right.visit_postorder(visit_fn, depth + 1, data) == STOP:
            return STOP

        if visit_fn and visit_fn(self, depth, data) == STOP:
            return STOP

    def get_root(self):
        """Return the root element of this tree"""
        result = self
        while result.parent:
            result = result.parent

        return result

    # **Child Management**
    #
    # Methods for setting the children on this expression.  These take care of
    # making sure that the proper parent assignments also take place.

    def set_left(
        self, child: "BinaryTreeNode" = None, clear_old_child_parent=False
    ) -> "BinaryTreeNode":
        """Set the left node to the passed `child`"""
        if child == self:
            raise ValueError("nodes cannot be their own children")
        if self.left is not None and clear_old_child_parent:
            self.left.parent = None
        self.left = child
        if self.left:
            self.left.parent = self

        return self

    def set_right(
        self, child: "BinaryTreeNode" = None, clear_old_child_parent=False
    ) -> "BinaryTreeNode":
        """Set the right node to the passed `child`"""
        if child == self:
            raise ValueError("nodes cannot be their own children")
        if self.right is not None and clear_old_child_parent:
            self.right.parent = None
        self.right = child
        if self.right:
            self.right.parent = self

        return self

    def get_side(self, child):
        """Determine whether the given `child` is the left or right child of this
        node"""
        if child == self.left:
            return LEFT

        if child == self.right:
            return RIGHT

        raise ValueError("BinaryTreeNode.get_side: not a child of this node")

    def set_side(self, child, side):
        """Set a new `child` on the given `side`"""
        if side == LEFT:
            return self.set_left(child)

        if side == RIGHT:
            return self.set_right(child)

        raise ValueError("BinaryTreeNode.set_side: Invalid side")

    def get_children(self):
        """Get children as an array.  If there are two children, the first object will
        always represent the left child, and the second will represent the right."""
        result = []
        if self.left:
            result.append(self.left)

        if self.right:
            result.append(self.right)

        return result

    def get_sibling(self) -> Optional["BinaryTreeNode"]:
        """Get the sibling node of this node.  If there is no parent, or the node
        has no sibling, the return value will be None."""
        if not self.parent:
            return None

        if self.parent.left == self:
            return self.parent.right

        if self.parent.right == self:
            return self.parent.left

        return None


class BinarySearchTree(BinaryTreeNode):
    """A binary search tree by key"""

    def __init__(self, key: Union[str, int, float] = None, **kwargs):
        super(BinarySearchTree, self).__init__(**kwargs)
        self.key = key

    def clone(self) -> BinaryTreeNode:
        result = super(BinarySearchTree, self).clone()
        result.key = self.key
        return result

    def insert(self, key) -> BinaryTreeNode:
        """Insert a node in the tree with the specified key."""
        node = self.get_root()
        while node:
            if key > node.key:
                if not node.right:
                    node.set_right(BinarySearchTree(key))
                    break
                node = node.right
            elif key < node.key:
                if not node.left:
                    node.set_left(BinarySearchTree(key))
                    break

                node = node.left
            else:
                break
        return self

    def find(self, key) -> Optional[BinaryTreeNode]:
        """Find a node in the tree by its key and return it.  Return None if the key
        is not found in the tree."""
        node = self.get_root()
        while node:
            if key > node.key:
                if not node.right:
                    return None

                node = node.right
                continue

            if key < node.key:
                if not node.left:
                    return None

                node = node.left
                continue

            if key == node.key:
                return node

            return None

        return None
